/*
 * barber.c: This program implements the classical IPC problem
 *           of "sleeping barber problem".
 *
 * by Alexandre Carissimi
 * Apache Project, IMAG, Grenoble, France
 * April, 22 1998.
 * 
 * To execute (after source on a0setup.csh flavour):
 *
 * a0run barber -a0n <n> 
 *
 * This program is based on Tanenbaum's solution given on
 * "Modern Operating System". Prentice Hall International.
 *
 */
 
 
#include	<stdlib.h>
#include	<ath0.h>


#define		SEED	       234
#define		MAXRAND		 5

#define		CLOSED		 0
#define		OPEN		 1
#define		FULL		 2
#define		BARBERTAG	 0
#define		CHAIRS		 5
#define		MAXCLIENTS	20

#define		ROOT		 0

a0tPort		Port;
a0tMutex	mutex;
a0tSemaphore	customers,barbers;
int		Service;
int		Open=OPEN;
int		waiting=0;
int		Limit=0;

void	cut_hair()
{
	sleep(3);
}


a0tError  CustomerBarberShop(a0tBuffer *arg)
{	
        int	i,Sender;
	
        a0Unpack(arg, a0_INT, &Sender, 1);	
	
	a0LockMutex(&mutex);         /*mutual exclusion to 'waiting'           */
        if ( ++Limit > MAXCLIENTS) {
	   Open = CLOSED;
	   a0UnlockMutex(&mutex);
	   return;
        }
	
	if (waiting < CHAIRS) {	     /*if there is no free chairs leave barber */
	   waiting = waiting + 1;    /*increment waiting customers             */
           printf(" Clients waiting: ");
	   for (i=0; i < waiting ; i++)
	       printf(" @:) ");
	   printf("%d\n", Sender);
	   a0VSemaphore(&customers); /*wake up barber if necessary             */
	   a0UnlockMutex(&mutex);    /*leaves critical region                  */
	   a0PSemaphore(&barbers);   /*wait for barber if it's not free        */
	                             /*now barber is cutting customer's hair   */
	}
	else {
	   --Limit;
	   a0UnlockMutex(&mutex);    /*barber shop is full... come back later  */
	   printf("Customer %d found barber shop full...\n", Sender);
	}  
	 
	return;
}


void	Barber(void)
{
	int	   i;
	a0tRequest Request;

	printf("\n\n");	
	printf("############################################\n");
	printf("#                                          #\n");
	printf("#        OPENING BARBER SHOP!!             #\n");
	printf("#                                          #\n");
	printf("############################################\n");
	    
	while ( Open == OPEN ){
	   if (waiting)
	      printf("Next please...\n");
	   else
	      printf("Wait for client...\n");
	      
	   a0PSemaphore(&customers);   /* waits for clients. Sleep if none   */
	   a0LockMutex(&mutex);        /* mutual exclusion for 'waiting'     */
	   waiting = waiting - 1;      /* take a customer to cut hair        */	
           printf(" Clients waiting: ");
	   for (i=0; i < waiting ; i++)
	       printf(" @:) ");
	   printf("    Barber is working... %%< %%< %%< %%<  \n");	      
	   a0VSemaphore(&barbers);     /* wake up barber                     */
	   a0UnlockMutex(&mutex);      /* leaves critical region             */
	   cut_hair();                 /* barber is working...               */	   
	}

        for( ; waiting > 0; waiting--) {
           printf(" Closing...Clients waiting: ");
	   for (i=0; i < waiting ; i++)
	       printf(" @:) ");
	   printf("    Barber is working... %%< %%< %%< %%<  \n");	      
	   a0VSemaphore(&barbers);     /* wake up barber                     */
	   cut_hair();                 /* barber is working...               */	   
	}   
	  			 
	for (i=1; i < a0NodeCount; i++)
	    a0Send(&Port, i, BARBERTAG, &Request, a0_INT, &Open, 1);

	printf("############################################\n");
	printf("#        CLOSING BARBER SHOP!!             #\n");
	printf("############################################\n");
		    
	return;
}


/****************************************************************************
                       SLAVES SIDE: Customers creation
 ****************************************************************************/
 
 
void Customer_Creation(void)
{
	int	   TimeWait;	/*time between customers creation            */
	int        result=0;    /*asynchronous communication is done or not  */
	int	   Client;
	a0tRequest ReqComm;     /*request used for communication             */
	a0tBuffer  Buffer;
		
	srand(SEED + a0SelfNode);
	        
       /*
        *  it will receive the moment when barbershop is closing. No
	* more customers will be created. Barber sends Open==CLOSED
	*/
	 
	a0IReceive(&Port, ROOT, BARBERTAG, &ReqComm, a0_INT, &Open, 1);
 
       /*
        * While barber shop is open it wait between 1 to 5 seconds to
	* generate a customer. Customers are generated until barber side
	* sends a message saying that barbershop is closed.
	*/
	
	Client = a0SelfNode;
	result = FALSE;	
	a0NewBuffer(&Buffer, A0ThreadBufferType, sizeof(int));
	
	result = 0;
	while (result==0) {

	   TimeWait = rand()%MAXRAND;
	   sleep(TimeWait==0?1:TimeWait);
	   
	   a0Pack(&Buffer, a0_INT, &Client, 1);
	   a0IStartRemoteThread(ROOT, Service, A0ServiceScheduling,
	                       A0ServicePriority, A0ServiceStack,
			       NULL, &Buffer);
		     
	   a0ClearBuffer(&Buffer);
	   Client +=10;	
	   result=0;	       
	   a0TestRequest(&ReqComm, &result);
	}

	return;
}


/****************************************************************************
                          	MAIN
 ****************************************************************************/
 
int	main(int argc, char *argv[])
{

	a0Init(&argc, &argv);
	a0NewPort(&Port, 1, A0GlobalCreation);
	a0NewService(&Service, CustomerBarberShop, A0DefaultScheduling,
	             A0DefaultPriority, A0DefaultStack);
        a0InitCommit();
	
	if (a0NodeCount < 2) {
	   printf("#\n# It must be run at least with TWO nodes!\n#\n");
	   a0Abort(0);
	}
	   
	   	
	if (a0SelfNode == ROOT)  {
	   if (a0NewMutex(&mutex) || 
	       a0NewSemaphore(&customers, 0) ||
	       a0NewSemaphore(&barbers, 0)) {
	       printf("#\n#\nError on synchronization variables initilization!\n");
	       a0Abort(0);
	   }
	   Barber();
	}
	else
	   Customer_Creation();
	
	a0DisposePort(&Port);
	a0Terminate();
}
	
			     
	 

