/*
 * amandel.c: This program evaluates mandelbrot set on a parallel way.
 *            It can be possible to execute on as a synthetic program
 *            changing some options. It uses Athapascan.
 *
 * by Alexandre Carissimi
 * Apache Project, IMAG, Grenoble, France
 * REVIEWED 25, january, 1999.
 *
 * Compiling:
 *	1. %source $A0HOMEDIR/bin/a0setup.csh
 *      2. make amandel
 *
 * Executing:
 *	%a0run amandel -a0n n [-a0verbose] [-iter iter] [-nt nt]
 *            [-image image] [-region region] [-trace file]
 *
 * Using the options:
 *      a0n	: no. of nodes (no default is set)
 *      iter    : no. iteration to test mandelbrot set (100)
 *      image   : image size to be considered (256)
 *      region  : region size to be considered (32)
 *      trace   : trace file name (trace)
 *      nt	: no. threads (1)
 *
 * The default values are given inside parentesis.
 *
 */
 
#include	<sys/time.h>
#include	<stdio.h>
#include	"parser.h"
#include	<ath0.h>
#include	<unistd.h>
#include	<errno.h>


/*************************************
   DATATYPE DEFINITIONS AND CONSTANTS
 *************************************/

 
#define 	WORKTAG			0
#define		RESULTAG		1
#define		ROOT			0

typedef	struct	{
	double	r;
	double	i;
}complex;


int		REGION_SIZE = 32;
int		IMAGE_SIZE = 256;
int		MAX_ITERATIONS = 100;
int		SYNTH = 0;
int		NT = 1;

a0tPort		Port;		
double		scale_x;
double		scale_y;
double		translate_x;
double		translate_y;


/*********************************************************************
                        GENERAL POURPOSES ROUTINES
 *********************************************************************/
  
double GetTime(void)
{
   struct timeval time;
   double Time;

   gettimeofday(&time, (struct timezone *) NULL);
   Time = ((double) time.tv_sec)*1000000.0 + time.tv_usec;
   return(Time);
}


void	HelpLine(void)
{
   printf("Usage:\n");
   printf("a0run mmandel -a0n 1 [-a0verbose] [-iter iter] [-on] [-synth]");
   printf(" [-nt nt]\n");
   printf("   [-reps reps] [-image image] [-region region] [-trace file]\n");
   
   printf("    on	: initializes Athapascan (off)\n");
   printf("    synt	: evaluates mandel as a synthetic program (off)\n");
   printf("    reps     : no. execution to compute an average (100)\n");
   printf("    iter     : no. iteration to test mandelbrot set (100)\n");
   printf("    image    : image size to be considered (256)\n");
   printf("    region   : region size to be considered (32)\n");
   printf("    trace    : trace file name (trace)\n");
   printf("    nt       : no. threads (1)\n");
   printf("\n    Default values inside parentesis.\n");
}



/*********************************************************************
                          MASTER SECTION
 *********************************************************************/
 

static int harvest(char *grid)
{	
	int		len0, len1;		/* x,y dim region length */
	int		i;
	int		source;			/* who sent last message */
	char		*pgrid;			/* ptr within grid buffer */
	char		*region;
	char		*aux;
	int		result_dim[4];		/* result dimensions */	
	a0tRequest	Request;
	a0tRequestData  Data;
	a0tError	a0Error;

	a0tBuffer	Buffer;

        a0NewBuffer(&Buffer, A0SendBufferType, sizeof(int)*4 + REGION_SIZE*REGION_SIZE);
	
	region=(char *)malloc(sizeof(char)*REGION_SIZE*REGION_SIZE);
	if (region==NULL) {
	   printf("Region allocation memory error...\n");
	   a0Abort(0);
	}
	
	if (a0ReceiveBuffer(&Port, A0AnySource, RESULTAG, &Request, &Buffer))
	   a0Abort(0);	

	a0Unpack(&Buffer, a0_INT, result_dim, 4);	
	a0Unpack(&Buffer, a0_CHAR, region, REGION_SIZE*REGION_SIZE);	

	a0GetRequestData(&Request, &Data);	 
   	source=Data.source;
	
	len0 = result_dim[2] - result_dim[0] + 1;
	len1 = result_dim[3] - result_dim[1] + 1;
	pgrid = grid + (IMAGE_SIZE * result_dim[1] + result_dim[0]);
	aux = region;
	
	for (i = 0; i < len1; ++i) {
		memcpy(pgrid, aux, len0);
		aux += len0;
		pgrid += IMAGE_SIZE;
	}
           		     
	free(region);
	return(source);
}


static void seed(int rank, int x, int y)
{
   static       int index=0;
   int		work[4];	        /* network message desc. */
   a0tRequest	Request;

   work[0] = x;		
   work[1] = y;
   work[2] = (x + REGION_SIZE > IMAGE_SIZE) ?
		  IMAGE_SIZE - 1 : x + REGION_SIZE - 1;
   work[3] = (y + REGION_SIZE > IMAGE_SIZE) ?
		  IMAGE_SIZE - 1 : y + REGION_SIZE - 1;
		  
       
   printf("%4d master: allocating block (%2d, %2d) - (%2d, %2d) to process %d\n",
		index++, work[0], work[1], work[2], work[3], rank);
   						
   if (a0Send(&Port, rank, WORKTAG, &Request, a0_INT, work, 4))
       a0Abort(0); 
   return;
}


/*********************************************************************
                          SLAVE SECTION
 *********************************************************************/
  

static char xy2color(double x, double y)
{
        char            colour;
        int             i;
        complex         z, c;
        double          mag2;
        double          scale;

        scale = ((double) 255) / MAX_ITERATIONS;
        colour = -1;

        z.r = z.i = 0;
        c.r = x;
        c.i = y;

        for (i = 0; i < MAX_ITERATIONS; ++i) {

                x = (z.r * z.r) - (z.i * z.i) + c.r;
                y = (2 * z.r * z.i) + c.i;

                z.r = x;
                z.i = y;

                mag2 = (z.r * z.r) + (z.i * z.i);

                if (mag2 >= 4) {
                   colour = (char) ((i * scale) + 0.5);
                   break;
                }
        }

        return(colour);
}



a0tError SlaveMandel(void *arg)
{
   int		work[4];
   int		lrx, ulx, lry, uly, region_size, i, j;
   char		*region, *aux; 
   a0tRequest	Request;
   a0tBuffer	Buffer;
   
   a0NewBuffer(&Buffer, A0SendBufferType, sizeof(int)*4 + REGION_SIZE*REGION_SIZE);

   for(;;) {
   
      if (a0Receive(&Port, ROOT, WORKTAG, &Request, 
                    a0_INT, work, 4) != A0ErrOk)
         a0Abort(0); 
            		     
      if (work[0] == 9999)
          break;
      
      ulx = work[0];
      uly = work[1];
      lrx = work[2];
      lry = work[3];
            
      region_size = (lrx - ulx + 1)*(lry - uly + 1);
      region = (char *) malloc (sizeof(char) * region_size);
      
      if (region==NULL) {
         printf("Memory allocation error on region...\n");
      	 a0Abort(0);
      }
	
      aux = region; 
      for(i = uly ; i <= lry ; i++) 
         for(j =ulx ; j <=lrx; j ++) 
	    *aux++=xy2color(translate_x + scale_x*j,
	                    translate_y + scale_y*i);
			    
      a0Pack(&Buffer, a0_INT, work, 4);
      a0Pack(&Buffer, a0_CHAR, region, region_size);
      if (a0SendBuffer(&Port, ROOT, RESULTAG, &Request, &Buffer))
         a0Abort(0);	
            		     	 		    
      a0ClearBuffer(&Buffer);
   }
	
   a0ExitThread(0);	   
}



/*-----------------------------------------------------------------------*

 *-----------------------------------------------------------------------*/


main(int argc, char *argv[])

{
	double		grid_coord[4];		/* complex grid description */
	int		i, n, nx, ny;
	int		ii, nt, TotalWork;
	int		rank;			/* compute task's rank      */
	int		x, y;			/* pixel coordinates        */
	char		*grid, *TestName;
	double		t0,t1,tt;
	FILE		*fp,*fptrace;	        /* output ppm file          */ 
	char		*TraceFile="trace";	
	a0tThread	*ThreadId;
	a0tError	a0Error;
	a0tRequest	Request;

   a0Init(&argc,&argv);
   a0NewPort(&Port, 3, A0GlobalCreation);
   a0InitCommit();
	
   if (a0NodeCount < 2) {
       printf("# It must have at most 2 nodes...\n");
       a0Abort(0);
   }
   
 /*
  * Hardcode the universe. Default values.
  */
 
   grid_coord[0] = -2.0;      /*x upper left corner    */
   grid_coord[1] = 1.25;      /*y upper left corner    */
   grid_coord[2] = 0.5;       /*x1 bottom left corner  */
   grid_coord[3] = -1.25;     /*y1 bottom left corner  */
        
 /*
  * Parsers command line arguments
  */
  
   if (ArgGetDoubleList(argc,argv,"-coord",4,grid_coord)==-1) {
      HelpLine();
      exit(0);
   }

   if (( grid_coord[2] < grid_coord[0] ) ||
       ( grid_coord[1] < grid_coord[3] )) {
       printf("Upper left coordinates must be greater than lower right\n");
       exit(0);
   }    

   if ( ArgGetInt(argc,argv,"-region", &REGION_SIZE)==-1) {
      printf("Region error\n");HelpLine();
      a0Abort(0);
   }
 
   if ( ArgGetInt(argc,argv,"-image", &IMAGE_SIZE)==-1) {
      printf("Image error\n");HelpLine();
      a0Abort(0);
   }
   
   if ( ArgGetInt(argc,argv,"-nt", &NT)==-1) {
      printf("Nt error\n");HelpLine();
      a0Abort(0);
   }

   if ( IMAGE_SIZE % REGION_SIZE ) {
      printf("REGION Size must be multiple of IMAGE size!! \n");
      HelpLine();a0Abort(0);
   }
  
   if ( ArgGetInt(argc,argv,"-iter", &MAX_ITERATIONS)==-1) {
      printf("Iter error\n");HelpLine();
      a0Abort(0);
   }
   
       
   if ( ArgGetStr(argc, argv, "-trace", &TraceFile)==-1) {
      HelpLine();
      a0Abort(0);
   }
   
   TestName="Mandelbrot";

   nx = IMAGE_SIZE/REGION_SIZE;
   ny = IMAGE_SIZE/REGION_SIZE;
			
   scale_x =  (grid_coord[2] - grid_coord[0]) / IMAGE_SIZE;
   scale_y = -((grid_coord[1] - grid_coord[3]) /IMAGE_SIZE);
   translate_x = grid_coord[0];
   translate_y = grid_coord[1];

  /*
   *  Traditional master slave model. Here we do this separation
   *  the slaves creates new threads to execute mandel evaluation
   *  the root does work distribution and control end of task
   */
 
 
   if ( a0SelfNode !=ROOT) {
      ThreadId = (a0tThread *) malloc (sizeof(a0tThread)*NT);
      if (ThreadId==NULL) {
         printf("Error on thread id memory allocation...\n");
         a0Abort(0);
      }
      
      for ( i = 0; i < NT ; i++)
  	 a0Error=a0NewSlave(&ThreadId[i], A0DefaultScheduling,
  			    A0DefaultPriority, A0DefaultStack,
  			    SlaveMandel, (void *)NULL);
 
      for ( i = 0; i < NT ; i++)
  	   a0JoinSlave(&ThreadId[i], NULL);
      a0DisposePort(&Port);
      a0Terminate();
      exit(0);
   }

   printf("#\n#    %s Evaluation (Athapascan)\n", TestName);
   printf("#\n# region (%4.2f,%4.2f)(%4.2f,%4.2f)\n#\n",
   	   grid_coord[0], grid_coord[1], grid_coord[2], grid_coord[3]);
   printf("# image: %d  region: %d  threads: %d  iter: %d\n",
   	     IMAGE_SIZE, REGION_SIZE, NT, MAX_ITERATIONS);
	     
   grid = (char *)malloc(sizeof(char)*IMAGE_SIZE*IMAGE_SIZE);
   if (grid == NULL ) {
      printf("Memory allocation error on image malloc...\n");
      a0Abort(0);
   }
	
			
   /*
    * Distribute first work order to all slaves.
    */
 
    tt=0;
        
        y = 0;
        n = 1;
        nt = 1;
	TotalWork = 0;
        t0 = GetTime();	
	
	while (( y < IMAGE_SIZE) && (TotalWork < (a0NodeCount-1)*NT) && ( n < a0NodeCount)) {
	  for (x = 0; (x < IMAGE_SIZE) && ( TotalWork < (a0NodeCount-1)*NT );x += REGION_SIZE){

	      if ( n > (a0NodeCount-1))
	          n=1;
	  		 
	      if ( TotalWork < (a0NodeCount-1)*NT ) {
	          seed(n, x, y);
	          TotalWork++;
		  n++;
	      }
	      
	      if (( n > a0NodeCount) && (( TotalWork % NT ) == 0))
	              y += REGION_SIZE;		      	       
	   }	       	    
	}
        	
		  
        for (; (y < IMAGE_SIZE); y += REGION_SIZE) { 
   	    for (; x < IMAGE_SIZE; x += REGION_SIZE) {
   		 n = harvest(grid);
   		 seed(n, x, y);
   	    }
	    x = 0;
	}
     
       /*
        * Wait for the last results
        */

        for (i = 0; i < TotalWork; i++)
   	        harvest(grid);
 
        t1 = GetTime();
	
        tt = (t1-t0);
   
    
  /*
   * Computation is finished, so we can terminate all the
   * slaves
   */
   
    n = 9999;
    for (rank = 1; rank < a0NodeCount; ++rank)
   	for (i = 0; i < NT ; i++) {
   	   a0Send(&Port, rank, WORKTAG, &Request, a0_INT, &n, 4);
	}   
	    
    printf("\n Execution Time: %10.4f msec\n\n", tt/1000.0);
      
    fptrace = NULL;
   
    if ((fptrace = fopen(TraceFile, "r")) == NULL) {
      fptrace = fopen(TraceFile, "w+");
      if ( fptrace == NULL) {
         printf("Trace file open error...\n");
         exit(0);
      }
      fprintf(fptrace, "#\n#      %s Multithreaded Trace\n", TestName);
      fprintf(fptrace, "#\n# Trace file: %s\n", TraceFile);
      fprintf(fptrace, "#\n# Image  Region  Thread  Temps (msec)");
      fprintf(fptrace, "  Iter. \n");
    }
    else {
        fptrace = fopen(TraceFile, "a+");
    }
        
    if ( fptrace != NULL ) {
      fprintf(fptrace, "%6d  %6d  %6d  %10.4f %5d \n",
              IMAGE_SIZE, REGION_SIZE, NT, tt/1000.0, MAX_ITERATIONS);
      fclose(fptrace);
    }

   
   /*
    *	Makes a portable pix map file. P3 is a magic number
    *	followed by image size, pixel depth, and a set of
    *	RGB values one for each pixel
    */
 
    if (SYNTH==0) {
        printf("################################################\n");
        printf("#\n# Please, wait!! I'm writing image file...\n#\n");
        printf("################################################\n"); 

        fp = fopen ("mandel.ppm","w");
        if (fp != NULL) {
           fprintf(fp,"P3\n");
           fprintf(fp,"%d %d\n",IMAGE_SIZE, IMAGE_SIZE);
           fprintf(fp,"7\n");

           for(n=0; n < IMAGE_SIZE*IMAGE_SIZE; n++)
              fprintf(fp,"%d %d %d \n", (grid[n]>>5)&6,
                     (grid[n]>>3)&7, (grid[n])&7);
           fprintf(fp,"\n");
           fclose(fp);
        }
        else
           printf("Error: file open ppm\n");
    }      
      
    printf("master: done.\n");
    
   free(grid);   
   a0DisposePort(&Port);
   a0Terminate();
}

