/*
 *
 * a0buffer.c : Functions for the Ath0 buffers
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * May 28, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include "a0error.h"
#include "a0buffer.h"
#include "a0mpi.h"


/* debugging macro */
#define a0Debug_onX(a) printf a
#define a0Debug(a)


/**************************************************************************/
/* Clears a buffer (as if it was just created) */
/**************************************************************************/

extern a0tError _a0ClearBuffer(a0tBuffer *buffer)
{
	long dummy_long = 0;	/* Just for packing nothing */
	int dummy_int = 0;	/* Just for packing nothing */


	buffer->position = 0;	/* no packs yet */

	/* We pack dummy athapascan data to make room for the real data that
	 * will come when the user has finished packing all its data in the buffer.
	 * The number and type of such packs depend on the type of buffer used.
	 */
	switch (buffer->bufftype) {
	case A0SendBufferType:
		break;
	case A0UrgentBufferType:
		_a0Pack(buffer, a0_INT, &dummy_int, 1);		/* service # */
		break;
	case A0ThreadBufferType:
		_a0Pack(buffer, a0_INT, &dummy_int, 1);		/* service # */
		_a0Pack(buffer, a0_INT, &dummy_int, 1);		/* sched */
		_a0Pack(buffer, a0_INT, &dummy_int, 1);		/* prio */
		_a0Pack(buffer, a0_UNSIGNED_LONG, &dummy_long, 1);	/* stack size */
		break;
	default:
		return A0ErrInvBuffType;
	}
	buffer->userdata = buffer->position;
	return A0ErrOk;
}


/**************************************************************************/
/* Allocates memory and creates a new buffer. */
/**************************************************************************/

extern a0tError _a0NewBuffer(a0tBuffer *buffer,
			     a0tBufferType bufftype,
			     size_t size)
{
	int size1, size2;

	/* Compute the number of bytes needed for athapascan zone,
	depending on
	 * buffer type.
	 */
	switch (bufftype) {
	case A0SendBufferType:
		size1 = size2 = 0;
		break;
	case A0UrgentBufferType:
		akMPI_Pack_size(1, MPI_INT, MPI_COMM_WORLD, &size1);
		size2 = 0;
		break;
	case A0ThreadBufferType:
		akMPI_Pack_size(3, MPI_INT, MPI_COMM_WORLD, &size1);
		akMPI_Pack_size(1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD,
		&size2);
		break;
	default:
		return A0ErrInvBuffType;
	}

	/* We have an upper bound on the number of bytes athapascan will
	need at
	 * the beginning of the buffer.
	 */
	if (!(buffer->storage = (void *) malloc(size + size1 + size2)))
		return A0ErrNoMem;

	buffer->bufftype = bufftype;
	buffer->size = size + size1 + size2;

	return _a0ClearBuffer(buffer);
}


/**************************************************************************/
/* Creates a new buffer with preallocated memory */
/**************************************************************************/

extern a0tError _a0NewBufferStorage(a0tBuffer *buffer,
				    a0tBufferType bufftype,
				    size_t size,
				    void *storage)
{
	buffer->storage = storage;
	buffer->bufftype = bufftype;
	buffer->size = size;

	return _a0ClearBuffer(buffer);
}


/**************************************************************************/
/* Deletes a buffer (allocated by _a0NewBuffer) and frees its storage. */
/**************************************************************************/

extern a0tError _a0DisposeBuffer(a0tBuffer *buffer)
{
	if (!buffer->storage)
		return A0ErrInvalid;

	free((void *) buffer->storage);
	buffer->storage = 0;

	return A0ErrOk;
}


/**************************************************************************/
/* Packs into a buffer. */
/**************************************************************************/

extern a0tError _a0Pack(a0tBuffer *buffer, a0tDatatype datatype,
			void *inbuf, int incount)
{
	int mpi_size, mpi_pos, r;

	/* MPI size & pos are ints */
	mpi_size = buffer->size;
	mpi_pos = buffer->position;

	r = akMPI_Pack(inbuf, incount, (MPI_Datatype) datatype, buffer->storage,
		       mpi_size, &mpi_pos, MPI_COMM_WORLD);
	buffer->position = mpi_pos;

	return r ? A0ErrFailPack : A0ErrOk;
}


/**************************************************************************/
/* Unpacks from a buffer. */
/**************************************************************************/

extern a0tError _a0Unpack(a0tBuffer *buffer, a0tDatatype datatype, 
			  void *outbuf, int outcount)
{
	int mpi_size, mpi_pos, r;

	/* MPI size & pos are ints */
	mpi_size = buffer->size;
	mpi_pos = buffer->position;

	r = akMPI_Unpack(buffer->storage, mpi_size, &mpi_pos, outbuf,
		      outcount, (MPI_Datatype) datatype, MPI_COMM_WORLD);
	buffer->position = mpi_pos;

	return r ? A0ErrFailPack : A0ErrOk;
}


/**************************************************************************
	Calculates the size of packed data
**************************************************************************/

extern a0tError _a0GetPackedSize(a0tDatatype datatype, size_t *size)
{
	int mpi_size;

	if (akMPI_Pack_size(1, datatype, MPI_COMM_WORLD, &mpi_size))
		return A0ErrInvalid;
	*size = mpi_size;
	return A0ErrOk;
}


/**************************************************************************
	Allocate a buffer with its descriptor
**************************************************************************/

extern a0tError _a0AllocateBuffer(a0tBuffer ** bufferpp,
			      a0tBufferType bufftype, size_t size)
{
	/* this function is used in athapascan internal functions, which
	   allocates a buffer with its descriptor. It is used to de-allocate
	   automatically with the callback function _a0DeallocateBuffer. */

	a0tError err;

	*bufferpp = (a0tBuffer *) malloc(sizeof(a0tBuffer));
	if (!*bufferpp)
		return A0ErrNoMem;

	err = _a0NewBuffer(*bufferpp, bufftype, size);
	if (err) {
		free(*bufferpp);
		return A0ErrNoMem;
	}
	return A0ErrOk;
}


/**************************************************************************
	Deallocate a buffer with its descriptor
**************************************************************************/

extern a0tError _a0DeallocateBuffer(a0tBuffer *bufferp)
{
	/* this function is used in athapascan internal functions, which
	   deallocates a buffer with its descriptor. It is used to de-allocate
	   automatically as a callback function. */

	_a0DisposeBuffer(bufferp);
	free(bufferp);
	return A0ErrOk;
}
