/*
 *
 * a0specific.c : Functions for the thread specific data.
 *
 * by Ilan Ginzburg, Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 17, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <string.h>

#include <akernel.h>
#include "a0error.h"
#include "a0specific.h"

/* debugging macro */
#define a0DebugXXX(a) printf a
#define a0Debug(a)


/*****************************************************************************/

/* Array containing the thread specific data, for service threads */
a0tThreadSpecific _a0Specific[A0MAXSERVICETHREADS];

/* Head of free list of thread specific data */
a0tThreadSpecific *_a0SpecificFreeHead;

/* Mutex protecting _a0SpecificFreeHead (and the rest of the list...) */
akpthread_mutex_t _a0SpecificMutex;

#ifdef A0_REUSETHR
/* a list of used but free structures for each priority */
a0tThreadSpecific *_a0SpecificPrioFreeHead[16];

/* Mutex protecting _a0SpecificPrioFreeHead (and the rest of the list...) */
akpthread_mutex_t _a0SpecificPrioMutex;
#endif

/* Key value for per-thread context */
akpthread_key_t _a0ThreadspecificKey;


/*****************************************************************************
	_ath0.breadSpecificDestructor - thread specific destructor
*****************************************************************************/

void _a0ThreadspecificDestructor(void *ctx_value) {
  /* The destruction is done elsewhere.
     Maybe some day we'll use this instead. */
}


/*****************************************************************************
	_a0InitSpecificData - init specific module
*****************************************************************************/

extern a0tError _a0InitSpecificData() {

  /* Initialise the thread specific structure and the thread
     specific key value. (to be called only once in every node) */

  int i;
  a0tThreadSpecific *main_thread;

  /* Create _a0SpecificMutex mutex. We don't lock it for the initialization
   * work because nobody else even THINKS of using it as we execute these
   * lines...
   */
  if (akpthread_mutex_init(&_a0SpecificMutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
	_a0Error(A0ErrMutexInit, "specific");

  /* Create the free list */
  for (i=0; i<A0MAXSERVICETHREADS; i++) {
    _a0Specific[i].nextfree=_a0Specific+i+1;
#ifdef A0_REUSETHR
    _a0Specific[i].flags &= ~A0SpecificUsed;
#endif
  }
  /* correct last link */
  _a0Specific[A0MAXSERVICETHREADS-1].nextfree = NULL;

  /* Head of free list */
  _a0SpecificFreeHead = _a0Specific;

#ifdef A0_REUSETHR
  /* empty used structures lists */
  for (i=0; i<16; i++) 
    _a0SpecificPrioFreeHead[i] = 0;

  /* create the mutex to protect access */
  if (akpthread_mutex_init(&_a0SpecificPrioMutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
	_a0Error(A0ErrMutexInit, "specific/prio");
#endif

  /* Create the key used for thread-specific storage.
   * We don't use the destructor but the documentation I have on pthreads
   * doesn't say if we can pass a NULL instead...
   */
  if (akpthread_key_create(&_a0ThreadspecificKey, _a0ThreadspecificDestructor))
	_a0Error(A0ErrSpecific, "key create");

  /* We now create the thread specific information for the 'main' thread.
   * This thread is a bit special (compared to a slave or a service thread).
   */
  /* Get a storage place in the array */
  _a0NewSpecificData(&main_thread);
  /* Initialize the needed fields. The other fields does not have meaning
   * in the main() thread. Actually I think A0SpecificBuffer shouldn't have
   * a meaning either... :-)
   */
  /* No buffer attached to main() thread */
  main_thread->flags &= ~A0SpecificBuffer;

#ifdef A0_TRACE
  akTraceNewThreadId(&main_thread->logical_id);
  akTraceSetThreadId(main_thread->logical_id);
#endif /* A0_TRACE */

  /* Set the key for the main thread (the main thread is the thread executing
   * this code).
   */
  akpthread_setspecific(_a0ThreadspecificKey, (void *) main_thread);

  return A0ErrOk;
}


/*****************************************************************************
	_a0NewSpecificData - Allocate a thread specific data structure
*****************************************************************************/

extern a0tError _a0NewSpecificData(a0tThreadSpecific **data) {

  /* Does NOT initialize the 'returned' structure.
     Return the adress in data. */
  int i;

  a0Debug(("_a0NewSpecificData: starting\n"));

  /* Lock the list */
  a0ThreadsBlockedLock++;
  akpthread_mutex_lock(&_a0SpecificMutex);
  a0ThreadsBlockedLock--;

  if (!_a0SpecificFreeHead) {
    /* No free structures... */
    a0tThreadSpecific *more;
    more = (a0tThreadSpecific *) malloc(A0MAXSERVICETHREADS * sizeof(a0tThreadSpecific));
    if (!more) {
      *data = NULL;
      return A0ErrNoMem;
    }

    /* Create the free list */
    for (i=0; i<A0MAXSERVICETHREADS; i++) {
      more[i].nextfree = more + i + 1;
#ifdef A0_REUSETHR
      more[i].flags &= ~A0SpecificUsed;
#endif
    }
    more[A0MAXSERVICETHREADS-1].nextfree = NULL;
    _a0SpecificFreeHead = more;
  }

  /* get the first one on the free list... */
  *data = _a0SpecificFreeHead;
  /* ...and remove it from there */
  _a0SpecificFreeHead = _a0SpecificFreeHead->nextfree;

  /* Unlock the list */
  akpthread_mutex_unlock(&_a0SpecificMutex);

  return A0ErrOk;
}


/*****************************************************************************
	_a0FreeSpecificData - Free the pointed to specific thread data
*****************************************************************************/

extern a0tError _a0FreeSpecificData(a0tThreadSpecific *data) {

  /* Free the pointed to specific thread data, in the
     specific thread data structure (where else?...) */

  /* Lock the list */
  a0ThreadsBlockedLock++;
  akpthread_mutex_lock(&_a0SpecificMutex);
  a0ThreadsBlockedLock--;

  /* Insert our data at the head of the list */
  data->nextfree=_a0SpecificFreeHead;
  _a0SpecificFreeHead=data;

  /* Unlock the list */
  akpthread_mutex_unlock(&_a0SpecificMutex);

  return A0ErrOk;
}


/*****************************************************************************
	_a0GetSpecificData - Get the thread specific structure
*****************************************************************************/

extern a0tError _a0GetSpecificData(a0tThreadSpecific **data) {

  /* Get a pointer on the thread specific structure
     for the calling thread. */

  int rc;
  void *value;

  if (akpthread_getspecific_np(_a0ThreadspecificKey, &value))
	_a0Error(A0ErrSpecific, "not get");

  *data = (a0tThreadSpecific *) value;

  return A0ErrOk;
}


/*****************************************************************************
	_a0SetSpecificData - Set the thread specific structure
*****************************************************************************/

extern a0tError _a0SetSpecificData(a0tThreadSpecific *data) {

  /* Set a pointer on the thread specific structure
     for the calling thread. */

  if (akpthread_setspecific(_a0ThreadspecificKey, data))
	_a0Error(A0ErrSpecific, "not set");

  return A0ErrOk;
}


#ifdef A0_REUSETHR
/*****************************************************************************
	_a0NewSpecificPrio - gets a new structure or a used but free one
*****************************************************************************/

extern a0tError _a0NewSpecificPrio(a0tThreadSpecific **data, int prio) {

  a0Debug(("_a0NewSpecificPrio: starting\n"));

  /* if prio must be inherited, find my prio and give it to new thread */
  if (prio == A0InheritPriority) {
    a0tThreadSpecific *spec;

    _a0GetSpecificData(&spec);
    prio = spec->Prio;
  }

  /* Lock the list */
  a0ThreadsBlockedLock++;
  akpthread_mutex_lock(&_a0SpecificPrioMutex);
  a0ThreadsBlockedLock--;

  /* hard coded: only 16 different priorities. */
  if (prio>15) prio = 15;
  if (prio<0) prio = 0;

  if (!_a0SpecificPrioFreeHead[prio]) {
    /* No free structures... */
    akpthread_mutex_unlock(&_a0SpecificPrioMutex);
    return A0ErrNoMem;
  }

  /* get the first one on the free list... */
  *data = _a0SpecificPrioFreeHead[prio];

  /* ...and remove it from there */
  _a0SpecificPrioFreeHead[prio] = _a0SpecificPrioFreeHead[prio]->nextfree;

  /* mark structure as used */
  (*data)->flags = A0SpecificUsed;

  /* Unlock the list */
  akpthread_mutex_unlock(&_a0SpecificPrioMutex);

  return A0ErrOk;
}


/*****************************************************************************
	_a0FreeSpecificPrio - frees a structure already used
*****************************************************************************/

extern a0tError _a0FreeSpecificPrio(a0tThreadSpecific *data, int prio) {

  /* Lock the list */
  a0ThreadsBlockedLock++;
  akpthread_mutex_lock(&_a0SpecificPrioMutex);
  a0ThreadsBlockedLock--;

  /* hard coded: only 16 different priorities. */
  if (prio>15) prio = 15;
  if (prio<0) prio = 0;

  /* Insert our data at the head of the list */
  data->nextfree = _a0SpecificPrioFreeHead[prio];
  _a0SpecificPrioFreeHead[prio] = data;

  /* Unlock the list */
  akpthread_mutex_unlock(&_a0SpecificPrioMutex);

  return A0ErrOk;
}
#endif /* A0_REUSETHR */
