/*
 *
 * a0flock.c : a0 server Functions of DMA lock
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Oct 18, 1996
 *
 */

#define A0_SOURCE
#include <ath0.h>

#define A0F_SOURCE
#include "a0fmem.h"
#include "a0flock.h"
#include "a0flockd.h"
#include "a0fread.h"
#include "a0fwrite.h"


#define a0Debug_on(a) printf a
#define a0Debug(a)


/******************local variables & definitions*****************/

typedef struct _a0tLocalLock {
  int		type;
  int		hash;
  a0tMemRequest	answer;
  void		*start;
  void		*end;
  struct _a0tLocalLock *wait_head;
  struct _a0tLocalLock **wait_tail;
  struct _a0tLocalLock *next;
} a0tLocalLock;


#define A0LockHashSize 13

static a0tLocalLock *_a0LockHash[A0LockHashSize];
static a0tMutex _a0LockHashMutex;


/*************************************************************************
	_a0EvaluateHash -
*************************************************************************/

static void _a0EvaluateHash(a0tLocalLock *new) {

  unsigned char *p;
  int t = 0, n;

  p = (unsigned char *) &new->answer.region;
  for (n=0; n<sizeof(a0tDMARegion); n++) t += *p++;
  new->hash = t % A0LockHashSize;

  a0Debug(("_a0EvaluateHash %x hash=%d\n", new, new->hash));
}


/*************************************************************************
	_a0CheckLockHash -
*************************************************************************/

static a0tError _a0CheckLockHash(a0tLocalLock *new, int *collision) {

  a0tLocalLock *lock;

  a0Debug(("_a0CheckLockHash %x BEGIN rep=%d\n", new, new->answer.reply_node));

  *collision = 0;

  if (_a0LockMutex(&_a0LockHashMutex)) _a0Error(A0ErrMutexLock, "chash");

  lock = _a0LockHash[new->hash];

  /* If the end of both regions are greater than the begin of the
     other, then the regions have some intersection. */

  while (lock) {
    if ((lock->end > new->start) && (new->end > lock->start)) {
      /* insert in collision waiting queue */
      *(lock->wait_tail) = new;
      lock->wait_tail = &(new->next);
      new->next = 0;
      *collision = 1;
      break;
    }
    lock = lock->next;
  }
  if (!*collision) {
    /* insert in hash table */
    new->next = _a0LockHash[new->hash];
    _a0LockHash[new->hash] = new;
    new->wait_tail = &(new->wait_head);
    new->wait_head = NULL;
  }

  if (_a0UnlockMutex(&_a0LockHashMutex)) _a0Error(A0ErrMutexUnlock, "chash");

  a0Debug(("_a0CheckLockHash END %x collision=%d\n", new, *collision));

  return A0ErrOk;
}


/*************************************************************************
	_a0CheckReplyLock -
*************************************************************************/

static a0tError _a0CheckReplyLock(a0tLocalLock *lock) {

  a0tMemoryAddress address;
  a0tError ret = A0ErrOk;
  int collision;

  _a0CheckLockHash(lock, &collision);

  if (!collision) {
    /* no collision: can answer */
    /* if read-at-lock, send data */
    if (lock->type & A0ReadRequest)
      _a0ReplyRead(&lock->answer);

    /* send lock address */
    address.address = lock;
    ret = _a0Send(&_a0MemReplyPort, lock->answer.reply_node,
			lock->answer.reply_tag,
			NULL, a0_ADDRESS, &address, 1);
  }
  a0Debug(("_a0CheckReplyLock %s granted lock=%x rep=%d tag=%d data=%x\n",
  		(collision?"NOT":""),
		lock, lock->answer.reply_node, lock->answer.reply_tag,
		lock->answer.region.data.address));
  return ret;
}


/*************************************************************************
	_a0ExtractLockHash -
*************************************************************************/

static a0tError _a0ExtractLockHash(a0tLocalLock *old) {

  a0tLocalLock **lock, *next;

  a0Debug(("_a0ExtractLockHash %x BEGIN rep=%d\n", old, old->answer.reply_node));

  if (_a0LockMutex(&_a0LockHashMutex)) _a0Error(A0ErrMutexLock, "xhash");

  lock = &_a0LockHash[old->hash];
  while (*lock && (*lock != old)) lock = &(*lock)->next;

  /* check for a bug: lock should be in the list */
  if (!*lock) _a0Error(A0ErrFatal, "*lock xhash");
  *lock = old->next;

  if (_a0UnlockMutex(&_a0LockHashMutex)) _a0Error(A0ErrMutexUnlock, "xhash");

  /* Try to insert in the hash table all requests of the waiting queue. */
  old = old->wait_head;
  while (old) {
    next = old->next;
    _a0CheckReplyLock(old);
    old = next;
  }

  a0Debug(("_a0ExtractLockHash %x END\n", old));
  return A0ErrOk;
}


/*************************************************************************
	_a0ReplyLock
*************************************************************************/

extern a0tError _a0ReplyLock(a0tMemRequest *message) {

  a0tLocalLock *newlock;

  newlock = (a0tLocalLock *) malloc(sizeof(a0tLocalLock));
  if (!newlock) _a0Error(A0ErrMalloc, "lock");

  newlock->type = message->type;
  newlock->answer.region = message->region;
  newlock->answer.region_start = message->region_start;
  newlock->answer.region_size = message->region_size;
  newlock->answer.reply_node = message->reply_node;
  newlock->answer.reply_tag = message->reply_tag;

  a0Debug(("req lock %x type=%d reg=%x st=%d sz=%d rep=%d tag=%d\n",
		newlock,
		newlock->type, newlock->answer.region.data.address,
		newlock->answer.region_start, newlock->answer.region_size,
		newlock->answer.reply_node, newlock->answer.reply_tag));

  /* calculate addresses */
  if (_a0GetFormatOffset(newlock->answer.region.format.address,
			 newlock->answer.region.data.address,
			 newlock->answer.region_start,
			 &newlock->start))
    _a0Error(A0ErrDaemonMsg, "lock1off");
  if (_a0GetFormatOffset(newlock->answer.region.format.address,
			 newlock->answer.region.data.address,
			 newlock->answer.region_start + newlock->answer.region_size,
			 &newlock->end))
    _a0Error(A0ErrDaemonMsg, "lock2off");

  _a0EvaluateHash(newlock);
  _a0CheckReplyLock(newlock);

  return A0ErrOk;
}


/*************************************************************************
	_a0ReplyUnlock
*************************************************************************/

extern a0tError _a0ReplyUnlock(a0tMemRequest *message) {

  a0tMemoryAddress address;
  a0tLocalLock *lock;

  /* dirty stuff: to keep the mem request compatible, pass lock address
     and answer tag inside the region. region.data = lock address and
     region.node = reply_node tag. */
  lock = (a0tLocalLock *) message->region.data.address;
  lock->type = message->type;
  lock->answer.reply_tag = message->region.node;

  a0Debug(("req unlock %x type=%d reg=%x st=%d sz=%d rep=%d tag=%d\n",
		lock, lock->type, lock->answer.region.data.address,
		lock->answer.region_start, lock->answer.region_size,
		lock->answer.reply_node, lock->answer.reply_tag));

  if (lock->type & A0WriteRequest) {
    lock->answer.writer_node = lock->answer.reply_node;
    lock->answer.writer_tag = lock->answer.reply_tag;
    _a0ReplyWrite(&lock->answer);
  }
  else
    _a0Send(&_a0MemReplyPort, lock->answer.reply_node, lock->answer.reply_tag,
		NULL, a0_INT, &_a0MemDummy, 0);

  _a0ExtractLockHash(lock);

  return A0ErrOk;
}


/*************************************************************************
	_a0InitLockServer
*************************************************************************/

a0tError _a0InitLockServer(void) {

  a0tError ret;
  int i;

  /* create mutex to protect the Hash access */
  if (ret = _a0NewMutex(&_a0LockHashMutex))
    return ret;

  /* zero hash table */
  for (i=0; i<A0LockHashSize; i++)
    _a0LockHash[i] = NULL;

  return A0ErrOk;
}
