
/*
 * akcomm.c : Communication functions
 *
 * by Marcelo Pasin, Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 5, 1996
 *
 * Modified on november, 29, 1998 by CARISSIMI, Alexandre:
 * - Different wc list management;
 * - Critic sections were reduced. Now akMPI_Mutex protects only MPI
 *   calls, access to wc list has its own mutexes set;
 * - minors arrangement on code style and function calling sequence
 *  
 */

#define AKERNEL_SOURCE
#include "akthread.h"
#include "akconfig.h"
#include "akmpi.h"
#include "akcomm.h"
#include "akdaemon.h"
#include "akerror.h"
#include <sys/errno.h>
#include <unistd.h>


/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a)


#if defined (A0_TSOL2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__pthread_mutex_lock.__pthread_mutex_owner64==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#elif defined (A0_TLIN2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__m_lock.__status==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#else
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#endif

		     


int akNode;      /* Rank of the node */
int akSize;      /* Number of nodes in the computation */
int SpinValue=0; /* Number of spins before lock */


/* Flags to set communication daemon behaviour */

int	akTimeOut = 100000; /* 100ms: watch dog timer default            */
int	akTimePoll= 10000;  /*  10ms: time default to poll communication */ 
int	akTimed   = 0;      /* daemons runs without timed fashion        */  
int	akDaemonOn= 1;      /* communication daemon runs                 */
int	akNoComm  = 0;      /* communications are allowed                */
int	akAdjustTime = 0;   /* dynamic time out corretion is off         */


/* conters of blocked threads. one for each WC type + semaphore and mutex */

int akBlocked[akWCNTYPES + 2];

#ifdef A0_TRACE
/* Variable counting the total number of requests.
 * First call is returned 0, second 1 etc...
 * The numerotation reflects the ACTUAL order in which MPI has been called
 * (we guarantee this because as of the time of writing, all calls to MPI are
 * done in mutual exclusion, and we increment the counter and call MPI
 * atomically)
 */
static int akCounter[akWCNTYPES];
#endif /* A0_TRACE */

/*#define AK_COUNT*/
#ifdef AK_COUNT
static unsigned long akCountCommWait = 0;
static unsigned long akCountCommCompletedSend = 0;
static unsigned long akCountCommCompletedRecv = 0;
static unsigned long akCountCommCompletedProbe = 0;
static unsigned long akCountCommCompleted = 0;
static unsigned long akCountCommPostedSend = 0;
static unsigned long akCountCommPostedRecv = 0;
static unsigned long akCountCommPostedProbe = 0;
#endif




/*************************************************************************
		akInit - Starts the kernel up
 *************************************************************************/

extern int akInit(int *argcp, char ***argvp) {

  /* Temporary storage for the creation of akWakerUp thread */
  int n;

  akDebug(("Starting akInit...\n"));

  akpthread_init_np();
  
  n = sysconf(_SC_NPROCESSORS_ONLN);
  
  if (n>1)
  	SpinValue=100;
  else
  	SpinValue=0;

  /*
   *  Does MPI initialization and initializes also a global mutex
   *  to grant thread safety to MPI. Some functions on format
   *  library calls MPI in concurrence of communications daemons
   */
 
  if (akpthread_mutex_init(&akMPI_Mutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
     akError(akERRMUTEXINIT, "MPI");
 
  if (akpthread_mutex_init(&akDaemonRunMutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
     akError(akERRMUTEXINIT, "DAEMON");

        
  if (akMPI_Init(argcp, argvp))
     akError(akERRMPI, "init" );
  if (akMPI_Comm_rank(MPI_COMM_WORLD, &akNode))
     akError(akERRMPI, "rank" );
  if (akMPI_Comm_size(MPI_COMM_WORLD, &akSize))
     akError(akERRMPI, "size" );


#ifdef A0_TRACE
  /* get global clock initial time */
  a0_perf_define_initial_time ();
  /* Counting the number of send & receives */
  for (n=0; n<akWCNTYPES; n++) akCounter[n] = 0;
#endif

 /*
  * Number of threads blockeds on each type of WC: send, receive,
  * probe and other.
  */

  for (n=0; n<akWCNTYPES; n++) 
      akBlocked[n] = 0;

 /*
  * Initilizes request list structures
  */
  
  akInitWC();
  
 /*
  * Initialize thread daemon which will run polling function.
  * It will be scheduled at system basis fashion. Set Watch
  * dog threads (...just in case of)
  */

  akDebug(("Starting akDaemonInit...\n"));
  if (akDaemonInit()) 
     akError(akERRDAEMON, "init" );

  akDebug(("Done akInit...\n"));
  return 0;
}


/************************************************************************
		  akFinalize - Terminates the kernel
 ************************************************************************/

extern int akFinalize(void) 
{

  akDebug(("%d:akFinalize BEGIN\n", akNode));

  if (akDaemonTerminate()) 
     akError(akERRDAEMON, "terminate" );     
  if (akMPI_Finalize())
     akError(akERRMPI, "finalize" );

#ifdef AK_COUNT
  printf("node %d comm waits=%lu compl=%lu\n",
		akNode, akCountCommWait, akCountCommCompleted);
  printf("node %d comm posted send=%lu recv=%lu probe=%lu\n",
		akNode, akCountCommPostedSend,
		akCountCommPostedRecv,
		akCountCommPostedProbe);
  printf("node %d comm cmpl send=%lu recv=%lu probe=%lu\n",
		akNode, akCountCommCompletedSend,
		akCountCommCompletedRecv,
		akCountCommCompletedProbe);
#endif
  return 0;
}


/*************************************************************************
  * The Send and Receive functions do this sequence:
	- find out a wait context
	- ask MPI to communicate
	- if it's going to be waited (normally)
		initialise its semaphore
		store the handle address in the wait context,
			to the daemon put MPI status there
		store the wait context address in the handle,
			to the user wait or test on it later
	- put the w.c. in a state the daemon looks for it
*************************************************************************/


/******************************************************************

 ******************************************************************/
 
extern int akISend(void *buf, int count, MPI_Datatype datatype, int dest,
		   int tag, MPI_Comm comm, aktHandle *handlep,
		   void (*callback)(), void *callback_par) 
{
    aktWaitContext *wcp;
     
    akDebug(("%d:akISecv %d:%d %X\n", akNode, dest, tag, handlep));
    akDebug(("akISend: will MPI_ISend(%x,%d,%x,%d,%d,%x,%x)\n",
	     buf, count, datatype, dest, tag,comm, &(wcp->u.request)));
   /*
    * Looks for a unused wait contex, and after fill its fields
    * whit communication parameters
    */
     
    akWCAlloc(&wcp);
    wcp->flags |= akWCSEND;
    wcp->callback = callback;
    wcp->callback_par = callback_par;
    
   /*
    * MPI does the send. It returns a request identifier to be
    * tested or waited for. This request is stored in the w.c.
    */
    
    akDebug(("akISend: before MPI_Isend()\n"));    
    akpthread_mutex_lock(&akMPI_Mutex); 
    akDebug(("akISend: after mutex_lock()\n")); 
    if (MPI_Isend(buf, count, datatype, dest, tag,comm, &(wcp->u.request))){
	akpthread_mutex_unlock(&akMPI_Mutex);
	return akERRMPISEND;
    }
    akpthread_mutex_unlock(&akMPI_Mutex);
    akDebug(("akISend: after MPI_Isend()\n")); 
      
#ifdef AK_COUNT
    akCountCommPostedSend ++;
#endif

/*  Trace mode things */
 
#ifdef A0_TRACE  
  wcp->serial_no = akCounter[akWCRECV] ++;
  /* handlep can be NULL if request isn't going to be waited on */
  if (handlep) handlep->serial_no = wcp->serial_no;
#endif

#ifdef AK_DAENUL
    /* give only NULL handles to the daemon */
    if (handlep)
	  handlep->wc = wcp;
    else
	  akWCInsertList(wcp, handlep);
#else
    /* prepare the wc to be waited/tested on, give it to the daemon */
    akWCInsertList(wcp, handlep);
#endif

    akDebug(("%d:akISend %d:%d %X (%X) done\n",akNode,dest,tag,handlep,wcp));
#ifndef AKDAEMON_RUN
    akAdvance(1,NULL);
#endif
    return 0;
}

/*************************************************************************
         
 *************************************************************************/

extern int akSend(void *buf, int count, MPI_Datatype datatype, int dest,
		  int  tag, MPI_Comm comm, aktHandle *handlep) 
{
  int r;

  r = akISend(buf, count, datatype, dest, tag, comm, handlep, 0, 0);
  if (r) return r;
  return akWait(handlep);
}


/******************************************************************

 ******************************************************************/
 
extern int akIRecv(void *buf, int count, MPI_Datatype datatype, int source,
		   int tag, MPI_Comm comm, aktHandle *handlep,
		   void (*callback)(), void *callback_par) 
{
    aktWaitContext *wcp;
     
    akDebug(("%d:akIRecv %d:%d %X\n", akNode, source, tag, handlep));

   /*
    * Looks for a unused wait contex, and after fill its fields
    * whit communication parameters
    */
  
     
    akWCAlloc(&wcp);
    wcp->flags |= akWCRECV;
    wcp->callback = callback;
    wcp->callback_par = callback_par;
    
   /*
    * MPI does the receive. It returns a request identifier to be
    * tested or waited for. This request is stored in the w.c.
    */
    
    akpthread_mutex_lock(&akMPI_Mutex); 
    akDebug(("%d:akIRecv: will MPI_Irecv(%x,%d,%x,%d,%d,%x,%x)\n",akNode,
	     buf, count, datatype, source, tag,comm, &(wcp->u.request)));   
    if (MPI_Irecv(buf, count, datatype, source, tag,comm, &(wcp->u.request))){
	akpthread_mutex_unlock(&akMPI_Mutex);
	return akERRMPIRECV;
    }
    akpthread_mutex_unlock(&akMPI_Mutex);
    akDebug(("akIRecv: after MPI_mutex()\n")); 
      
#ifdef AK_COUNT
    akCountCommPostedRecv ++;
#endif

/*  Trace mode things */
 
#ifdef A0_TRACE  
  wcp->serial_no = akCounter[akWCRECV] ++;
  /* handlep can be NULL if request isn't going to be waited on */
  if (handlep) handlep->serial_no = wcp->serial_no;
#endif

#ifdef AK_DAENUL
    /* give only NULL handles to the daemon */
    if (handlep)
	  handlep->wc = wcp;
    else
	  akWCInsertList(wcp, handlep);
#else
    /* prepare the wc to be waited/tested on, give it to the daemon */
    akWCInsertList(wcp, handlep);
#endif

    akDebug(("%d:akRecv %d:%d %X (%X) done\n",akNode,source,tag,handlep,wcp));
#ifndef AKDAEMON_RUN
    akAdvance(1,NULL);
#endif
    return 0;
}


/*************************************************************************
          
 *************************************************************************/

extern int akRecv(void *buf, int count, MPI_Datatype datatype, int source,
		  int  tag, MPI_Comm comm, aktHandle *handlep) 
{
  int r;

  r = akIRecv(buf,count,datatype,source,tag,comm,handlep,0,0);
  if (r) return r;
  return akWait(handlep);
}


/*************************************************************************

 *************************************************************************/

extern int akIProbe(int source, int tag, MPI_Comm comm, aktHandle *handlep,
		    void (*callback)(), void *callback_par) 
{
    aktWaitContext	*wcp;
    int 		flag;

    akDebug(("akIProbe %d:%d comm=%d %X\n", source, tag, comm, handlep));

    /* Probe cannot have NULL handle */
    if (!handlep) return akERRNULLHANDLE;

#ifdef A0_TRACE
    /* Added here for the trace module */;
    handlep->serial_no = akCounter[akWCPROBE] ++;
#endif

    /* Well... Maybe just a nonblocking probe will do? */
  
  
    akpthread_mutex_lock(&akMPI_Mutex);
    if (MPI_Iprobe(source, tag, comm, &flag, &handlep->status)) {
       akpthread_mutex_unlock(&akMPI_Mutex);
       return akERRMPIRECV;
    }
    akpthread_mutex_unlock(&akMPI_Mutex);
    
#ifdef AK_COUNT
    akCountCommPostedProbe ++;
#endif

    if (flag) {
       /* God is great, the probe terminated */
       handlep->wc = 0;

#ifdef A0_TRACE
       akWCCallbackTrace(callback, callback_par, akWCPROBE,
			&(handlep->status), handlep->serial_no);
#else
       if (callback)
          (callback)(callback_par);
#endif

#ifdef AK_COUNT
       akCountCommCompletedProbe ++;
#endif

    return 0;
  }
   /*
    * Well... The probe didn't terminate, so now the daemon will redo it 
    * from time to time, until it terminates.
    *
    * Looks for a unused wait contex, and after fill its fields
    * whit communication parameters
    */
     
     akWCAlloc(&wcp);
     wcp->flags = akWCPROBE;
     wcp->callback = callback;
     wcp->callback_par = callback_par;


   /* 
    * Copy the information to the wait context so the daemon knows what
    * to probe 
    */
    
     wcp->u.probe.source = source;
     wcp->u.probe.tag = tag;
     wcp->u.probe.comm = comm;

#ifdef A0_TRACE
     /* Added here for the trace module */;
     wcp->serial_no = handlep->serial_no;
#endif

#ifdef AK_DAENUL
     /* give only NULL handles to the daemon */
     if (handlep)
	handlep->wc = wcp;
     else
	akWCInsertList(wcp, handlep);
#else
     /* prepare the wc to be waited/tested on, give it to the daemon */
     akWCInsertList(wcp, handlep);
#endif

  akDebug(("akIProbe %d:%d %X (%X) done\n", source, tag, handlep, wcp));
#ifndef AKDAEMON_RUN
    akAdvance(1,NULL);
#endif
  return 0;
}


/************************************************************************

 *************************************************************************/

extern int akProbe(int source, int tag, MPI_Comm comm, aktHandle *handlep) 
{
    int	r;

    r = akIProbe(source, tag, comm, handlep, 0, 0);
    if (r) return r;
    return akWait(handlep);
}



/*************************************************************************

 *************************************************************************/

extern int akWait(aktHandle *handlep) 
{
     int 	i, r = 0, f;

    /*
     * Its is possible to wait only non NULL handlers. When an handle
     * has already been waited on, we return (no error)
     */
     
     if (!handlep) 
        return akERRNULLHANDLE;	 
	 
     if (!handlep->wc) 
        return 0;

     akDebug(("%d:akWait %X (%X)\n", akNode, handlep, handlep->wc));

    /* count blocked threads */
     akBlocked[handlep->wc->flags & akWCTYPE] ++;

#ifdef AK_BLKMPI
    /* if MPI is thread-safe, block inside of it */
     switch (handlep->wc->flags&akWCTYPE) {
     case akWCSEND:
     case akWCRECV:
#ifndef AK_DAENUL
          akWCMoveToInUse3(handlep->wc);
#endif
          MPI_Wait(&(handlep->wc->u.request), &(handlep->status));
          break;
     case akWCPROBE:
#ifndef AK_DAENUL
          akWCMoveToInUse3(handlep->wc);
#endif
          MPI_Probe(handlep->wc->u.probe.source, handlep->wc->u.probe.tag,
	            handlep->wc->u.probe.comm, &(handlep->status));
          break;
     default:
#endif /* AK_BLKMPI */


    /* check if the request is ended. it always happens for short msgs */

/*
 *     if (!(handlep->wc->flags & akWCDONE))
 *      akWCCheck(handlep->wc);
 */
    
#ifdef AK_COUNT
    if (handlep->wc->flags & akWCDONE) {
       switch (handlep->wc->flags & akWCTYPE) {
         case akWCPROBE:   akCountCommCompletedProbe ++;   break;
         case akWCRECV:    akCountCommCompletedRecv ++;    break;
         case akWCSEND:    akCountCommCompletedSend ++;    break;
       }
       akCountCommCompleted ++;
    }
    akCountCommWait ++;
#endif

   /*
    * Blocks to wait completion. The daemon unlocks the semaphore
    * when it recognizes the completion. 
    */
    
    akpthread_mutex_lock(&(handlep->wc->MutexCond));
    while (!(handlep->wc->flags & akWCDONE))
     if(akpthread_cond_wait(&(handlep->wc->cond), &(handlep->wc->MutexCond))<0){
        akpthread_mutex_unlock(&(handlep->wc->MutexCond));
	akError(akERRSEMAP, "p wait");
     }
#ifdef AK_BLKMPI
  }
#endif /* AK_BLKMPI */

    /* count blocked threads */
  
    akBlocked[handlep->wc->flags & akWCTYPE] --;

    if (handlep->wc->flags & akCANCEL) 
       r = akERRCANCEL;

    /* clean up and release the wait context */
 
    akpthread_mutex_unlock(&(handlep->wc->MutexCond));

    akWCMoveToFree(handlep->wc);

    akDebug(("%d:akWait %X (%X) done\n", akNode, handlep, handlep->wc));
    handlep->wc = 0;
    return r;
}


/*************************************************************************

 *************************************************************************/

extern int akTest(aktHandle *handlep, int *result) 
{ 
    int 	r = 0;
    
    /*
     * Its is possible to wait only non NULL handlers. When an handle
     * has already been waited on, we return (no error)
     */
     
     if (!handlep) 
        return akERRNULLHANDLE;	 
	 
     if (!handlep->wc)  {
        *result = TRUE;
        return 0;
     }
  
     if (!(handlep->wc->flags & akWCDONE)) {
        /*akpthread_mutex_lock(&akDaemonRunMutex);*/
        akWCCheck(handlep->wc);
	/*akpthread_mutex_unlock(&akDaemonRunMutex);*/
     }
     
     if (handlep->wc->flags & akCANCEL)
        r = akERRCANCEL;
     if (*result = handlep->wc->flags & akWCDONE) {
#ifdef AK_COUNT
        switch (handlep->wc->flags & akWCTYPE) {
           case akWCPROBE:     akCountCommCompletedProbe ++;   break;
           case akWCRECV:      akCountCommCompletedRecv ++;    break;
           case akWCSEND:      akCountCommCompletedSend ++;    break;
        }
        akCountCommCompleted ++;
#endif
        akWCMoveToFree(handlep->wc);
        handlep->wc = 0; /* zero for subsequent waits or tests */
    }
  
    return r;
}



/*************************************************************************

 *************************************************************************/

extern int akWaitAny(aktHandle *array[], int size, int *index) 
{
   akpthread_cond_t	cond;
   int 			i, r, ct;

   r = akTestAny(array, size, index);

   if (*index == akNOCOMPLETED) {
      /* create cond variable to sleep waiting on it */
      if (akpthread_cond_init(&cond,akPTHREAD_CONDATTR_DEFAULT_NP))
	  akError(akERRCONDINIT, "wait-any");
      if (akpthread_mutex_lock(&akMPI_Mutex))
	  akError(akERRMUTEXLOCK, "wait-any");
      /* point all wc->any to cond */
      for(i = ct = 0; i<size; i++)
         if (array[i]) {
	    akDebug(("akWaitAny on %x (%d)\n", array[i], i));
#ifdef AK_DAENUL
	   akWCInsertList(array[i]->wc, array[i]);
#endif
	   array[i]->wc->any = &cond;
	   array[i]->wc->flags |= akWCANY;
	   ct ++;
         }
	 
	 
      if (!ct) {
         /* user passed a null table */
         if (akpthread_mutex_unlock(&akMPI_Mutex))
	    akError(akERRMUTEXLOCK, "wait-any-ct");
         akDebug(("akWaitAny all done\n"));
         *index = akNOCOMPLETED;
         return 0;
      }
           
     while (*index == akNOCOMPLETED) {
       akpthread_cond_wait(&cond, &akMPI_Mutex);
       for(i=0; i<size; i++) {
	  akDebug(("akWaitAny test %x (%d)\n", array[i], i));
	  if (array[i] && (array[i]->wc->flags & akWCDONE)) {
	    akDebug(("akWaitAny completed %x (%d)\n", array[i], i));
	    *index = i;
	    break;
	  }
       }
     }
     
     
    /* clear all wc->any */
    for(i=0; i<size; i++)
      if (array[i]) {
	array[i]->wc->flags &= ~akWCANY;
#ifdef AK_DAENUL
	akWCMoveToInUse3(array[i]->wc);
#endif
      };
    if (akpthread_mutex_unlock(&akMPI_Mutex))
        akError(akERRMUTEXLOCK, "wait-any-ct");
    if (akpthread_cond_destroy(&cond))
	akError(akERRCONDDESTR, "wait-any");
  }

  return r;
}


/*************************************************************************

 *************************************************************************/

extern int akTestAny(aktHandle *array[], int size, int *index) 
{
     int	i, succ = 0, r;

     for (i=0; i<size; i++) {
       if (array[i]) {
         r = akTest(array[i], &succ);
         if (r) {
	   *index = i;
	   return r;
         }
	 
         /* if one succeeded, returns the counter as index */
         if (succ) {
	   *index = i;
	   return 0;
         }
       }
    }

    /* no one completed */
     *index = akNOCOMPLETED;
     
     return 0;
}


/*************************************************************************

 *************************************************************************/

extern int akBlock(aktHandle *handlep,void (*callback)(),void *callback_par)
{
     aktWaitContext	*wcp;

     akDebug(("akBlock\n"));

     if (!handlep) 
        return akERRNULLHANDLE;

   /*
    * Looks for a unused wait contex, and after fill its fields
    * whit communication parameters
    */
     
     akWCAlloc(&wcp);
     wcp->flags |= akWCOTHER;
     wcp->callback = callback;
     wcp->callback_par = callback_par;

    /* remember handle address to the daemon store MPI status */
    
     wcp->statusp = &(handlep->status);
     
    /* remember w.c. address to the user wait/test on it */
    
     handlep->wc = wcp;

#ifdef A0_TRACE
    /* Added here for the trace module */;
     handlep->serial_no = wcp->serial_no = akCounter[akWCOTHER] ++;
#endif

     akDebug(("akBlock: (%X,%X) done\n", wcp, handlep));
     return 0;
}

/*************************************************************************

 *************************************************************************/

extern int akUnblock(aktHandle *handlep) 
{
    /*
     * Its is possible to unblock only non NULL handlers.
     */
     
     if (!handlep) 
        return akERRNULLHANDLE;	 

     if (!handlep->wc) 
        return akERRNULLHANDLE;

#ifdef AK_NSTOP
     akEverybodyBlocked = 0;
#endif

  akDebug(("akUnblock %x (%x)\n", handlep, handlep->wc));
  akWCUnblock(handlep->wc);
  akDebug(("akUnblock %x (%x) DONE\n", handlep, handlep->wc));
}


/*************************************************************************

 *************************************************************************/

extern int akCancel(aktHandle *handlep) 
{
    int	type;

    if (!handlep) 
       return akERRNULLHANDLE;
       
    if (!handlep->wc)
       return akERRNULLHANDLE;

    akDebug(("akCancel\n"));
    akWCRemoveList(handlep->wc);
    akWCUnblock(handlep->wc);
    handlep->wc->flags |= akCANCEL;
    type = handlep->wc->flags & akWCTYPE;
    if ((type == akWCSEND) || (type == akWCRECV))
      if (MPI_Cancel(&(handlep->wc->u.request)))
	  akError(akERRMPI, "cancel");

}


/*************************************************************************

 *************************************************************************/

extern int akNoWait(aktHandle *handlep) 
{
    int		r=0;

   /* 
    * This call is implemented in order to allow the higher level layer
    * to submit a non-NULL request and ask it to be treated as if a NULL
    * request had been posted. 
    */

    if (!handlep) 
       return akERRNULLHANDLE;

   /* When the handle has already been waited on, we return (no error) */
   
    if (!handlep->wc) 
       return 0;

     akDebug(("akNoWait (%X:%X)\n", handlep->wc, handlep->wc->u.request));

   /* 
    * If the daemon already completed the request,
    *	 clean up and release the wait context;
    * else
    *    mark the w.c. as 'no wait'. The daemon will free it.
    */
    
    if (handlep->wc->flags & akWCDONE)
       akWCMoveToFree(handlep->wc);
    else {
       (handlep->wc)->flags |= akWCNOWAIT;
       (handlep->wc)->statusp = &akDummyStatus;
    }

    akDebug(("akNoWait %X done\n", handlep->wc));

   /* We zero this for subsequent waits or tests */
   
    handlep->wc = 0;
    
    return r;
}

