/*
 * akdaemon.c : Kernel daemon version SMP/Mono
 *
 * by Alexandre CARISSIMI
 * Apache Project, IMAG, Grenoble, France
 * 29 November, 1998
 */


/* Kernel source file */

#define	AKERNEL_SOURCE
#include "akthread.h"
#include "akconfig.h"
#include "akmpi.h"
#include "akdaemon.h"
#include "akcomm.h"
#include "akwc.h"
#include "akerror.h"
#include <sys/errno.h>



/* debugging function */

#define akDebug_off(a) 
#define akDebug(a)  

#if defined(A0_TSOL2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__pthread_mutex_lock.__pthread_mutex_owner64==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#elif  defined(A0_TLIN2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__m_lock.__status==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#else
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#endif

		     
/*************************************************************************
	Global variables
 *************************************************************************/

#define akDAEMONNOTHING 0
#define akDAEMONWRAPPED 1
#define akDAEMONCOMPLETED 2

akpthread_mutex_t	akDaemonRunMutex;

static akpthread_t akWakerUpThread;	/* watch-dog thread id */
static akpthread_t akDaemonThread;	/* Daemon thread id */
static akpthread_attr_t akDaemonAttr;

static volatile int akDaemonRun;	/* daemon termination flag */
akpthread_cond_t  akWatchDogCond;	/* daemon sleeps waiting on it */


static volatile struct timespec akDaemonSleepDelta;

#ifdef AK_COUNT
static unsigned long akCountDaemonSleep = 0;
static unsigned long akCountDaemonTimedSleep = 0;
static unsigned long akCountDaemonYield = 0;
static unsigned long akCountDaemonAdvance = 0;
static unsigned long akCountDaemonTestSend = 0;
static unsigned long akCountDaemonTestRecv = 0;
static unsigned long akCountDaemonTestProbe = 0;
static unsigned long akCountDaemonChecked = 0;
static unsigned long akCountDaemonCompletedSend = 0;
static unsigned long akCountDaemonCompletedRecv = 0;
static unsigned long akCountDaemonCompletedProbe = 0;
static unsigned long akCountDaemonCompleted = 0;
static unsigned long akCountDaemonWrapped = 0;
static unsigned long akCountDaemonNothing = 0;
#endif

/*
 * This defines are related to Watch dog thread.
 */

#define		DONE	 	1
#define 	NOT_DONE	0
#define		MYTIMEOUT    	100000000    /*the time-out value*/
#define		ONE_SECOND  	1000000000   /*one second on nano seconds */

static	int WCScanning = NOT_DONE;

static int akPollingComm(void);
static void *akWatchDog(void* );
static void *akDaemonCommTimed(void* );
static void *akDaemonComm(void* );

/* the NEXT w.c. to be processed in the next loop of the daemon */

aktWaitContext *volatile akDaemonNext=NULL;


/******************************************************************************
 akAdvance : force a polling on MPI to check if some requests are ready.
 ******************************************************************************/
  
extern int akAdvance(void)
{
     int	r;
     
     if (akpthread_mutex_trylock(&akDaemonRunMutex)) 
        return akERRMUTEXLOCK;
	
     r=akPollingComm();
     WCScanning = DONE;
     akpthread_mutex_unlock(&akDaemonRunMutex);
  
     return akERROK;
}

/*****************************************************************************
 akPollingComm : checks all requests on wait context list.
 *****************************************************************************/
 
static int akPollingComm(void)
{
     aktWaitContext	*curr;
  
     do {  /* Get the head of request list */     
        spin_lock(akWCListMutex);
     
        if (akDaemonNext == NULL)
           akDaemonNext = akWCInUse2;
	
        curr = akDaemonNext;
        if (curr == NULL) { /* if all entries were checked it leaves */
           akpthread_mutex_unlock(&akWCListMutex);
	   return akDAEMONNOTHING;
         }
	 	
        akDaemonNext = akDaemonNext->next;
        akpthread_mutex_unlock(&akWCListMutex);              


#ifdef AK_COUNT
 	switch (curr->flags & akWCTYPE) {
 	  case akWCPROBE:  akCountDaemonTestProbe ++;	   break;
 	  case akWCRECV:   akCountDaemonTestRecv ++;	   break;
 	  case akWCSEND:   akCountDaemonTestSend ++;	   break;
 	}
#endif
	akDebug_off(("checking %X (flags=%X) akDaemonNext: %X \n",
 		 curr, curr->flags, akDaemonNext));
 	akWCCheck(curr);
    } while (0); 
}



/******************************************************************************
  akDaemonInit : all communications daemons
 ******************************************************************************/

extern int akDaemonInit(void) 
{
    akpthread_attr_t 		attr;
    struct aksched_param 	params;
    int 			sched;
    
    akDebug(("akDaemonInit BEGIN\n"));

    akDaemonRun = 1;
    akDaemonNext=NULL;

  /* We have three daemons responsible to check incoming messages.
   * The default one polls when it is scheduled on normal basis by
   * the thread scheduler. The second one, timed, becomes ready after
   * an elapsed time given on command line. The third one, is a watch
   * dog. It exist to grant a poll on a maximal time delay.
   */   
   
    if (akpthread_attr_init(&attr))
       akError(akERRDAEMON, "daemons attr");

    if (akpthread_attr_setscope(&attr, AK_DAEMON_SCOPE))
        akError(akERRDAEMON, "daemons scope");

    if (akpthread_attr_setinheritsched(&attr, akPTHREAD_EXPLICIT_SCHED))
       akError(akERRDAEMON, "daemons inherit");
      
    if (akpthread_attr_setschedpolicy(&attr, AK_DAEMON_SCHED))
        akError(akERRDAEMON, "daemons policy");
	
    params.aksched_priority = AK_DAEMON_PRIO;

#if !defined(A0_TLIN2) && !defined(A0_TAIX4)
   if (akpthread_attr_setschedparam(&attr, &params))
      akError(akERRDAEMON, "deamon prio");
#endif

#ifdef A0_TAIX4
   if (akpthread_attr_setdetachstate(&attr, akPTHREAD_CREATE_JOINABLE))
      akError(akERRDAEMON, "deamon detach");
#endif


    if (akNoComm == 1 ) {
       akDebug(("Option says no COMMUNICATION DEMON... so END\n"));
       akDaemonRun = 0;       
       if (akpthread_cond_init(&akWatchDogCond, akPTHREAD_CONDATTR_DEFAULT_NP))
          akError(akERRCONDINIT, "daemon WatchDog");  
       if (akpthread_create(&akWakerUpThread, &attr, akWatchDog, NULL))
          akError(akERRDAEMON, "start WatchDog");    
       if (akpthread_attr_destroy(&attr))
          akError(akERRDAEMON, "daemon attr destruction");	  
       return 0;  
    }
    
    akDebug(("akDaemonInit: will start daemon\n"));

    if (akTimed == 1) {
       if (akpthread_create(&akDaemonThread, &attr, akDaemonCommTimed, NULL))
    	  akError(akERRDAEMON, "start daemon");
    }
    else {
       if (akpthread_create(&akDaemonThread, &attr, akDaemonComm, NULL))
    	  akError(akERRDAEMON, "start daemon");
    }
   
    akDebug(("akDaemonInit: daemon started!\n"));
    
    if (akpthread_attr_destroy(&attr))
        akError(akERRDAEMON, "daemon attr destruction");
      
    akDebug(("akDaemonInit END\n"));

    return 0;
}


/*******************************************************************************
 akDaemonCommTimed : polls on a periodically fashion trigged by a time
                     interruption. This daemon is activate by -a0timed option
 ******************************************************************************/
 		     
static void *akDaemonCommTimed(void* arg)
{
     int		r;
     aktWaitContext	*curr;
     struct timespec    wakeup;
     struct timeval     tp;
     
     while(akDaemonRun) {
        if (akpthread_mutex_trylock(&akDaemonRunMutex)) {
	     continue;
	}
	 	 
	gettimeofday(&tp, NULL);
	wakeup.tv_nsec = tp.tv_usec*1000 + akTimePoll*1000;
	wakeup.tv_sec = tp.tv_sec;
	if ((wakeup.tv_nsec / ONE_SECOND) > 0) {
	   wakeup.tv_sec += (wakeup.tv_nsec / ONE_SECOND);
	   wakeup.tv_nsec -= ((wakeup.tv_nsec / ONE_SECOND)*ONE_SECOND);
	}
	
        akpthread_cond_timedwait(&akWatchDogCond,&akDaemonRunMutex,&wakeup);
        
	r=akPollingComm();	   
	WCScanning = DONE;
	akpthread_mutex_unlock(&akDaemonRunMutex);
     }
    
     akDebug(("%d: Leaving POLLING DAEMON\n",akNode));
    
#ifdef AK_CANCEL
    /* 
     * Cancel all the MPI requests, so it doesn't blame the ``abrupt''
     * termination. Ideally, only the final user requests should be
     * cancelled. Here, all daemons AND user requests will be cancelled.
     */

     ;
     akDebug(("%d: It will CANCEL all pending MPI requests\n",akNode));     

     
     akpthread_mutex_lock(&akWCListMutex);   
     curr = akWCInUse2;
     akDebug(("%d: CANCELING INITIATING BY...%X\n",akNode,curr));
     while (curr) {
       akDebug(("%d: Canceling...%X\n",akNode,curr));
       akpthread_mutex_lock(&akMPI_Mutex);
       if (((curr->flags&akWCTYPE)==akWCSEND)||
          ((curr->flags&akWCTYPE)==akWCRECV))  
          if (MPI_Cancel(&(curr->u.request))) {
	     akpthread_mutex_unlock(&akMPI_Mutex);
	     akError(akERRMPI, "cancel daemon");
	  };
       akpthread_mutex_unlock(&akMPI_Mutex);
       curr = curr->next;
     akDebug(("%d: After Canceling...next: %X\n",akNode, curr));
     }
     akpthread_mutex_unlock(&akWCListMutex);
     
     akDebug(("%d: End of CANCEL...\n",akNode));  
#endif
 
     akDebug(("%d: POLLING DAEMON Ciao\n",akNode)); 

} 

/******************************************************************************
 akWatchDog : grants a poll every "aktimeout" microseconds. Useful if user
              forgets akAdvance call on demand basis mode
 ******************************************************************************/

static void *akWatchDog(void *arg)
{

     int		r;
     struct timespec wakeup;
     struct timeval  tp;
       
     while(akDaemonRun){
        akpthread_mutex_lock(&akDaemonRunMutex);
	
	gettimeofday(&tp, NULL);
	wakeup.tv_nsec = tp.tv_usec*1000 + akTimeOut*1000;
	wakeup.tv_sec = tp.tv_sec;
	if ((wakeup.tv_nsec / ONE_SECOND) > 0) {
	   wakeup.tv_sec += (wakeup.tv_nsec / ONE_SECOND);
	   wakeup.tv_nsec -= ((wakeup.tv_nsec / ONE_SECOND)*ONE_SECOND);
	}
	
	WCScanning = NOT_DONE;
	     
        akpthread_cond_timedwait(&akWatchDogCond,&akDaemonRunMutex,&wakeup);
        if (WCScanning == NOT_DONE)
	   r=akPollingComm();
	   
	akpthread_mutex_unlock(&akDaemonRunMutex);
     }     

     akDebug(("%d: Leaving WATCH_DOG thread\n", akNode));
}
	        				
/*****************************************************************************
 akDaemonComm : the official communication daemon. It polls MPI on a system
                schedul basis. It is the default mode.
 *****************************************************************************/		

static void *akDaemonComm(void* arg)
{
     int		r;
     aktWaitContext	*curr;

     while(akDaemonRun) {
         if (akpthread_mutex_trylock(&akDaemonRunMutex)) {
	     aksched_yield();
	     continue;
	 }	 
 	 
         r=akPollingComm();
	 WCScanning = DONE;
	 akpthread_mutex_unlock(&akDaemonRunMutex);
	 aksched_yield();
    }
    
    akDebug(("%d: Leaving POLLING DAEMON\n",akNode));
    
#ifdef AK_CANCEL
    /* 
     * Cancel all the MPI requests, so it doesn't blame the ``abrupt''
     * termination. Ideally, only the final user requests should be
     * cancelled. Here, all daemons AND user requests will be cancelled.
     */

     ;
     akDebug(("%d: It will CANCEL all pending MPI requests\n",akNode));     

     
     akpthread_mutex_lock(&akWCListMutex);   
     curr = akWCInUse2;
     akDebug(("%d: CANCELING INITIATING BY...%X\n",akNode,curr));
     while (curr) {
       akDebug(("%d: Canceling...%X\n",akNode,curr));
       akpthread_mutex_lock(&akMPI_Mutex);
       if (((curr->flags&akWCTYPE)==akWCSEND)||
          ((curr->flags&akWCTYPE)==akWCRECV))  
          if (MPI_Cancel(&(curr->u.request))) {
	     akpthread_mutex_unlock(&akMPI_Mutex);
	     akError(akERRMPI, "cancel daemon");
	  };
       akpthread_mutex_unlock(&akMPI_Mutex);
       curr = curr->next;
     akDebug(("%d: After Canceling...next: %X\n",akNode, curr));
     }
     akpthread_mutex_unlock(&akWCListMutex);
     
     akDebug(("%d: End of CANCEL...\n",akNode));  
#endif
 
     akDebug(("%d: POLLING DAEMON Ciao\n",akNode)); 

} 



/*************************************************************************
 akDaemonTerminate : daemon termination
 *************************************************************************/

extern int akDaemonTerminate(void) 
{

   void *status;

   akDebug(("%d:akDaemonTerminate BEGIN\n", akNode));
  
   /*  tell the demon to stop in the next loop    */
  
   akDaemonRun = 0;

   /* just in case the daemon is sleeping */
   
   akDebug(("%d:akDaemonTerminate BEFORE cond_signal\n", akNode));
   
   akpthread_mutex_lock(&akDaemonRunMutex);
   WCScanning = DONE;
   akpthread_cond_signal(&akWatchDogCond);
   akpthread_mutex_unlock(&akDaemonRunMutex);

   akDebug(("%d:akDaemonTerminate BEFORE join watcher\n", akNode)); 
   if ((akNoComm == 0)  && (akDaemonOn==0))
      if (akpthread_join(akWakerUpThread, NULL))
         akError(akERRDAEMON, "Watch Dog join");

   akDebug(("%d:akDaemonTerminate AFTER join watcher\n", akNode));
  
   /* wait the daemon thread end */

#ifdef AKDAEMON_RUN   
   akDebug(("%d:akDaemonTerminate BEFORE join daemon\n", akNode)); 

   if ((akNoComm == 0) && (akDaemonOn == 1))     
      if (akpthread_join(akDaemonThread, NULL))
         akError(akERRDAEMON, "daemon join");
   akDebug(("%d:akDaemonTerminate after join daemon\n", akNode));
#else
   akDaemonComm(&status);
#endif

   akDebug(("%d:akDaemonTerminate after join watcher\n", akNode));
   akDebug(("%d:akDaemonTerminate END\n", akNode));

#ifdef AK_COUNT
   printf("node %d daemon completed=%lu wrapped=%lu nothing=%lu chk/wrap=%f\n",
		akNode, akCountDaemonCompleted, akCountDaemonWrapped,
		akCountDaemonNothing,
		(float) akCountDaemonChecked /
			(float) akCountDaemonWrapped);
   printf("node %d daemon sleep=%lu yield=%lu timed=%lu advance=%lu\n",
		akNode, akCountDaemonSleep, akCountDaemonYield,
		akCountDaemonTimedSleep, akCountDaemonAdvance);
   printf("node %d daemon test send=%lu recv=%lu probe=%lu\n",
		akNode, akCountDaemonTestSend, akCountDaemonTestRecv,
		akCountDaemonTestProbe);
   printf("node %d daemon cmpl send=%lu recv=%lu probe=%lu\n",
		akNode, akCountDaemonCompletedSend, akCountDaemonCompletedRecv,
		akCountDaemonCompletedProbe);
#endif

  return 0;
}
