
/*
 * akwc.c : Kernel wait context operations
 *
 * by Alexandre CARISSIMI
 * Apache Project, IMAG, Grenoble, France
 * November 29, 1998
 *
 */


/* Kernel source file */
#define	  AKERNEL_SOURCE
#include "akthread.h"
#include "akwc.h"
#include "akcomm.h"
#include "akdaemon.h"
#include "akerror.h"
#include <semaphore.h>
#include <errno.h>

/* debugging function */
#define akDebug_on(a) printf a
#define akDebug(a) 


#if defined(A0_TSOL2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__pthread_mutex_lock.__pthread_mutex_owner64==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#elif  defined(A0_TLIN2)
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
		        if (m.__m_lock.__status==0) \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#else
#define spin_lock(m) do {int spin;                                             \
                      for(spin=SpinValue; spin>0; spin--) {                    \
                           if (pthread_mutex_trylock(&m) != EBUSY)             \
                               break;                                          \
		      }                                                        \
		      if (spin==0) pthread_mutex_lock(&m);                     \
		     } while(0)
#endif

		
/* Wait context table */

aktWaitContext akWCTable[akMAXWC];

/* wait context lists heads */

aktWaitContext *volatile akWCFree = NULL;
aktWaitContext *volatile akWCInUse2 = NULL;


/* wait context InUse2 list counter */

int akWCInUse2Count = 0;

/* 
 * dummy handle to give to the daemon store the status, 
 * when a request is not expected to be waited or tested on
 */
 
 
MPI_Status akDummyStatus;

akpthread_mutex_t	akWCFreeMutex;
akpthread_mutex_t	akWCListMutex;


extern int akInitWC(void)
{

     akpthread_mutex_init(&akWCFreeMutex, NULL);
     akpthread_mutex_init(&akWCListMutex, NULL);
     
     akInsertWCTable(akWCTable, akMAXWC);
     
     return 0;
}     



extern int akInsertWCTable(aktWaitContext *tab, int size) 
{
     int	i;

    /*
     *  Link a vector elements as a list. Each element points to the next
     *  one. Also initializes flags (other fields are filled by calling
     *  threads and condition variable associated to each element.
     */

     tab[0].prev = NULL;
     
     for(i=0; i<size; i++) {
        tab[i].flags = akWCFREE; /* Other fields we don't care about */
        tab[i].next = tab+i+1; /* Last one will be improperly set */
        akpthread_cond_init(&tab[i].cond, akPTHREAD_CONDATTR_DEFAULT_NP);
     }
     
     tab[akMAXWC-1].next = NULL; /* correct last link */
     akWCFree = tab;  /* the free list begins in the first element */    
     return 0;
}


extern int akWCAlloc(aktWaitContext **wcpp)
{
     aktWaitContext	*wcp;
     aktWaitContext	*tab;
     
     if (akNoComm == 1)
        akError(akERRNOCOMM, "Error: You selected NO COMMUNICATION option!!");
     
     spin_lock(akWCFreeMutex);
     if (akWCFree == NULL) {
        tab = (aktWaitContext *) malloc (akMAXWC * sizeof(aktWaitContext));
        if (tab == NULL) 
	   akError(akERRNOWC, "Allocation error on WaitContext");
        akInsertWCTable(tab, akMAXWC);
     }

     wcp = akWCFree;
     akWCFree = akWCFree->next;
     wcp->flags = akWCINUSE1;
     akpthread_mutex_init(&(wcp->MutexCond),NULL);
     *wcpp = wcp;
     akpthread_mutex_unlock(&akWCFreeMutex);
     
     return 0;
}


extern int akWCInsertList(aktWaitContext *wcp, void *hp)
{
     aktHandle *handlep = (aktHandle *) hp;
         
    /*
     * sets this element on InUse2 state preserving the other flags
     */
     
     wcp->flags &= ~akWCSTATE; 
     wcp->flags |= akWCINUSE2;
     
    /*
     * This wc will be not waited, so when it will complete we can
     * free it directly. However MPI needs some place to store status
     * so we pass "dummy"
     */
     
     if ( handlep == NULL ) {
        wcp->flags |= akWCNOWAIT;
	wcp->statusp = &akDummyStatus;
     }
     else {
        wcp->statusp = &(handlep->status);
	handlep->wc = wcp;
     }

    /*
     * Finally we will insert this element on the request list
     */

     spin_lock(akWCListMutex);  
     wcp->next = akWCInUse2; 
     if (akWCInUse2 != NULL) 
        akWCInUse2->prev = wcp;
   	
     wcp->prev = NULL;      
     akWCInUse2 = wcp;
     akWCInUse2Count ++;	/* count one more w.c. */
     
     akpthread_mutex_unlock(&akWCListMutex);
          
     akDebug(("akWCMoveToInUse2 :Inserted wc=%X f=%x t=%s\n",
	    wcp, wcp->flags,
	    ((wcp->flags&akWCTYPE)==akWCSEND)?"send":
	    (((wcp->flags&akWCTYPE)==akWCRECV)?"recv":
	    (((wcp->flags&akWCTYPE)==akWCPROBE)?"probe":
	    "other(?)"))));

     return 0;
}

extern int akWCRemoveList(aktWaitContext *wcp)
{

   /*
    * security test: this node really belongs to request List?
    */
    
    if ((wcp->flags & akWCSTATE) != akWCINUSE2)
       return -1;
        
    spin_lock(akWCListMutex);

    if ( wcp == akDaemonNext )  
       akDaemonNext = wcp->next;
         
    if (wcp->next)
       wcp->next->prev = wcp->prev;
       
    if (wcp->prev)
        wcp->prev->next = wcp->next;
    else 
        akWCInUse2 = wcp->next;
	
    wcp->flags=(wcp->flags & ~akWCSTATE) | akWCINUSE3;
    akWCInUse2Count--;
    
    akpthread_mutex_unlock(&akWCListMutex);

}


extern int akWCMoveToFree(aktWaitContext *wcp) 
{

  /* 
   * inserts in the head of the Free list 
   */
   
   wcp->flags = akWCFREE;
      
   spin_lock(akWCFreeMutex);
   
   wcp->next = akWCFree;
   akWCFree = wcp;
   akpthread_mutex_destroy(&(wcp->MutexCond));

   akpthread_mutex_unlock(&akWCFreeMutex);
   
   return 0;
}


/*************************************************************************
akWCUnblock - complete and free the w.c., signal waiting thread
*************************************************************************/


extern int akWCUnblock(aktWaitContext *wcp)
{

  /*
   * if there is a callback function associeted to it, we execute it now
   */

  if (wcp->callback)
    (wcp->callback)(wcp->callback_par);

  /*
   * Is this request gonna be waited for or tested? if not you can free it,
   * otherwise we need to signal the waiting thread.
   */
   
   if (wcp->flags & akWCNOWAIT) 
      akWCMoveToFree(wcp);
  
   else {
      wcp->flags |= akWCDONE;
      /* check if it's a part of a wait/test any */
      if (wcp->flags & akWCANY) {
         akpthread_mutex_lock(&(wcp->MutexCond));
         if (akpthread_cond_signal(wcp->any)<0) {
	    akpthread_mutex_unlock(&(wcp->MutexCond));
	    akError(akERRSEMAP, "V any-daemon");
	 }
	 akpthread_mutex_unlock(&(wcp->MutexCond));
      }
      else {
         if ((wcp->flags&akWCTYPE)==akWCOTHER)
         akDebug(("akWCUnblock signal cond %x\n", &wcp->cond));
	 akpthread_mutex_lock(&(wcp->MutexCond));
         if (akpthread_cond_signal(&(wcp->cond))<0) {
	    akpthread_mutex_unlock(&(wcp->MutexCond));
	    akError(akERRSEMAP, "V daemon");
	 }
	 akpthread_mutex_unlock(&(wcp->MutexCond));
      }
   }
}


/***************************************************************************

 ***************************************************************************/
 
extern int akWCCheck(aktWaitContext *wcp)
{
     int	completed;

    /*
     * Scan Wait Context list
     */
  
     completed=0;
         
     switch (wcp->flags & akWCTYPE) {
        case akWCPROBE:
             akDebug(("akWCCheck %X %d:%d comm=%d\n", wcp, wcp->u.probe.source,
                    wcp->u.probe.tag, wcp->u.probe.comm));

	     spin_lock(akMPI_Mutex);
             if (MPI_Iprobe(wcp->u.probe.source, wcp->u.probe.tag,
		wcp->u.probe.comm, &completed, wcp->statusp))
		akError(akERRMPI, "test daemon2");
	     akpthread_mutex_unlock(&akMPI_Mutex);	     
             break;
	     
	case akWCSEND:
        case akWCRECV:
	     spin_lock(akMPI_Mutex);
             if (MPI_Test(&(wcp->u.request), &completed, wcp->statusp))
		akError(akERRMPI, "test daemon1");
	     akpthread_mutex_unlock(&akMPI_Mutex);
             break;
	     
        default:
             return;
     }

     if (completed) {
        akDebug(("completed wc=%X h=%x t=%s\n", wcp,
		(wcp->statusp==&akDummyStatus)? 0 : wcp->statusp,
		((wcp->flags&akWCTYPE)==akWCSEND)?"send":
		(((wcp->flags&akWCTYPE)==akWCRECV)?"recv":
		(((wcp->flags&akWCTYPE)==akWCPROBE)?"probe":
		"other(?)"))));

	akWCRemoveList(wcp);       
        akWCUnblock(wcp);
     }
     
     return(completed);
}




