/****************************************************************************/
/* File a0_perf_trace_buffer.h                                              */
/* Ph. Waille   05/07/96                                                    */
/****************************************************************************/

#ifndef _A0_PERF_TRACE_BUFFER_H
#define _A0_PERF_TRACE_BUFFER_H

#include "a0_perf_timing.h"
#include "a0semap.h"
#include <stdio.h>

#define A0_PERF_RELEASE "ATH_Perf_000.00"
#define A0_PERF_RELEASE_SIZE  16	

#define A0_NODE_SIZE	4
#define A0_PERF_TIME_SIZE	12


/* The Athapascan tracing system uses per thread, fixed size trace buffers.
Buffers are managed according to a producers/consumer policy.
Producers are user threads to trace and the consumer is a flushing daemon
to dump the buffers on the local file system of the node (in the local trace
file).  Each starting user thread is allocatted a trace buffer to record its
own events.  On buffer overflow, the full buffer is passed to the flushing
dameon and another free buffer is provided to the user thread as soon as
avalaible. Buffers flushed to the trace file returns to the free buffer list.*/

/* A buffer are passed to the flushing daemon
  + by its owner thread when full or during the clean up when the threads dies 
  + by the main thread, within the terminate primitive 
*/

/* Each buffer belongs to a global list of all buffers (in any state) */
/* Three others list are maintained :
  + full : buffers to be passed to the daemon
  + flushing : buffers currently handled by the daemon
  + free : free buffers (newly created or returned by the dameon)
*/

/* The writing pointer index always points to a record limit */

/* Put_full_buffer atomically performs two jobs : enqueue the buffer and
compute its content size (index - content). Race condition : put_perf_buffer
called twice for the same buffer by its owner thread and by the main thread
within the terminate primitive. If content size already <> 0 => do no enqueue */

/* Content [0] is the first available event record position */
/* 4 is meaningless. Its true size should be sizeof (buffer) - 5 * 4 */

/* Content size should be declared as a union of an array of bytes
(to store the size in the variable length format) and a long int to 
effciently reset it to the default initial value. Lazy programming : 
only declared as a long int */



#define TRACE_BUFFER_SIZE	2000
#define NB_TRACE_BUFFERS	50

#define MIN_TRACE_BUFFER_SIZE   200
#define MIN_NB_TRACE_BUFFERS 	3

#define DEFAULT_TRACE_FILE_NAME	"/TMP/perf_trace"

/* Non standard 0 value, in either little or big endian integers */
#define A0_PERF_INITIAL_NULL_SIZE	0x00808000

/* Internal trace buffer format */
/* Dummy content size : buffer size is dynamically defined by the user */
/* in a parameter in the command line */

typedef struct a0_perf_buffer_type {
	struct a0_perf_buffer_type *next;	/* buffers queue link */
	struct a0_perf_buffer_type *next_global;/* next in the global list */
	char *index;			/* writing pointer */
	char *high_water_mark;		/* last event record position allowed */
	unsigned long int size;		/* content size */
	/* Do not insert anything between size and content field */
	char content [4];		/* meaningless size of 4 */
	} a0_perf_trace_buffer;

/* External trace buffer format */
/* Same as above, without the first fields, meaningless after execution */

typedef struct a0_reader_buffer_type {
	unsigned long int size;		/* content size */
	/* Do not insert anything between size and content field */
	char content [4];		/* meaningless size of 4 */
	} a0_perf_reader_buffer;

#define FLUSHING_DAEMON_STACK	4000


extern a0tContext a0_perf_Trace_buffer_context;

extern a0_perf_trace_buffer *a0_perf_main_thread_buffer;

extern int trace_buffer_size;
extern int nb_trace_buffers;
extern char trace_file_name[];

extern FILE *fout;
extern int a0_perf_nb_wait_buffer;



extern a0_perf_trace_buffer *a0_perf_get_free_trace_buffer ();

extern void a0_perf_put_full_trace_buffer (a0_perf_trace_buffer *buffer);

extern void a0_perf_put_free_trace_buffers (a0_perf_trace_buffer *first,
                                            a0_perf_trace_buffer *last);

extern int a0_perf_get_full_trace_buffers ();

extern void a0_perf_init_trace_buffers (int argc,char **argv);

void a0_perf_init_free_list ();

extern void a0_perf_flush_trace_buffers ();

extern void a0_perf_wait_daemon_end ();

extern void a0_perf_start_flushing_daemon ();

#endif /* A0_PERF_TRACE_BUFFER_H */
