/*
 * akmpi.h : mpi stubs for thread safety.
 * Feb 3, 1996. Ilan Ginzburg, Marcelo Pasin.
 *
 * Athapascan Kernel.
 *
 * Modifications: (Alexandre Carissimi)
 *	- mutex exclusive to MPI functions
 */

 /* Eliminates multiple includes of this file */



#ifndef _AKMPI_H
#define _AKMPI_H

#undef WANT_MPI2CPP    /*to LAM implantation not consider C++ */
#undef HAVE_MPI_CPP    /*to MPICH implantation not consider C++ */
	
#include <pthread.h>
#include <mpi.h>
#include "akthread.h"


/* Mutex protecting accesses to MPI library */

extern akpthread_mutex_t akMPI_Mutex;


/* set akSETUP_MPI */

#if	defined(A0_LAM)
#define 		akSETUP_MPI "lam"
#elif	defined(A0_MPICH)
#define 		akSETUP_MPI "mpich"
#elif	defined(A0_MPIF)
#define 		akSETUP_MPI "mpif"
#elif	defined(A0_MPIBM)
#define 		akSETUP_MPI "ibm"
#elif defined(A0_MPISGI)
#define			akSETUP_MPI "sgi"
#endif


/*************************************************************************
 *
 * MPI functions
 *
 * We give higher layers the use of MPI non-communicating and non-
 * blocking functions (datatypes, packing...). Calls must be protected for
 * thread reentrance.
 *
 * The parameters for the functions here are the same as for the associated
 * MPI functions.
 *
 * The #defines of MPI can be used as-is.
 *
 */


int akMPI_Type_contiguous(int count, MPI_Datatype oldtype, 
			MPI_Datatype *newtype);
int akMPI_Type_vector(int count, int blocklength, int stride, 
		    MPI_Datatype oldtype, MPI_Datatype *newtype);
int akMPI_Type_hvector(int count, int blocklength, MPI_Aint stride, 
		     MPI_Datatype oldtype, MPI_Datatype *newtype);
int akMPI_Type_indexed(int count, int *array_of_blocklengths, 
		     int *array_of_displacements, MPI_Datatype oldtype, 
		     MPI_Datatype *newtype);
int akMPI_Type_hindexed(int count, int *array_of_blocklengths, 
		      MPI_Aint *array_of_displacements, MPI_Datatype oldtype, 
		      MPI_Datatype *newtype);
int akMPI_Type_struct(int count, int *array_of_blocklengths, 
		    MPI_Aint *array_of_displacements, 
		    MPI_Datatype *array_of_types, MPI_Datatype *newtype);
int akMPI_Address(void* location, MPI_Aint *address);
int akMPI_Type_extent(MPI_Datatype datatype, MPI_Aint *extent);
int akMPI_Type_size(MPI_Datatype datatype, int *size);
int akMPI_Type_lb(MPI_Datatype datatype, MPI_Aint* displacement);
int akMPI_Type_ub(MPI_Datatype datatype, MPI_Aint* displacement);
int akMPI_Type_commit(MPI_Datatype *datatype);
int akMPI_Type_free(MPI_Datatype *datatype);
int akMPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int *count);
int akMPI_Get_elements(MPI_Status *status, MPI_Datatype datatype, int *count);
int akMPI_Pack(void* inbuf, int incount, MPI_Datatype datatype, void *outbuf,
             int outsize, int *position,  MPI_Comm comm);
int akMPI_Unpack(void* inbuf, int insize, int *position, void *outbuf,
               int outcount, MPI_Datatype datatype, MPI_Comm comm);
int akMPI_Pack_size(int incount, MPI_Datatype datatype, MPI_Comm comm,
                  int *size);

int akMPI_Comm_size(MPI_Comm comm, int *size);
int akMPI_Comm_rank(MPI_Comm comm, int *rank);
int akMPI_Comm_compare(MPI_Comm comm1, MPI_Comm comm2, int *result);
int akMPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm);
int akMPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm);
int akMPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm);
int akMPI_Comm_free(MPI_Comm *comm);
int akMPI_Comm_test_inter(MPI_Comm comm, int *flag);
int akMPI_Comm_remote_size(MPI_Comm comm, int *size);
int akMPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group);
int akMPI_Abort(MPI_Comm comm, int errorcode);



#ifdef AKERNEL_SOURCE
/****************************************************************************
		Included only for compiling the kernel
****************************************************************************/


int akMPI_Init(int *argcp, char ***argvp);
int akMPI_Finalize(void);
int akMPI_Iprobe(int source, int tag, MPI_Comm comm, int *flag,
		MPI_Status *status);
/* Warning: the serial_no parameter in akMPI_Isend is used to give serial
   numbers to all the sends the akernel does, for use by the trace module.
   It is NOT an original MPI parameter to the send primitive. */
int akMPI_Isend(void* buf, int count, MPI_Datatype datatype, int dest, int tag,
		MPI_Comm comm, MPI_Request *request, long *serial_no);
int akMPI_Irecv(void* buf, int count, MPI_Datatype datatype, int src, int tag,
		MPI_Comm comm, MPI_Request *request, long *serial_no);
int akMPI_Test(MPI_Request *request, int *flag, MPI_Status *status);
int akMPI_Request_free(MPI_Request *request);
int akMPI_Cancel(MPI_Request *request);

#endif /* AKERNEL_SOURCE */

#endif /* _AKMPI_H */
