/*
 * a0init.c : Initialisation of athapascan0
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 21, 1996
 *
 */

/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>


/* don't know why it doesn't compile */
#if defined(A0_MPIF)
#define O_RDONLY 0000000
#define O_WRONLY 0000001
#define O_CREAT  0000400
#define O_TRUNC  0001000
#define S_IFIFO  0010000
#endif

#include "a0init.h"
#include "a0error.h"
#include "a0ports.h"
#include "a0servd.h"
#include "a0urgentd.h"
#include "a0services.h"
#include "a0buffer.h"
#include "a0thread.h"
#include "a0prio.h"
#include "a0mpi.h"
#include "a0semap.h"
#include "a0parser.h"


/* debugging macro */
#define a0Debug(a)
#define a0Debug_on(a) printf a

/* Rank of the local node in the computation */
int a0SelfNode;
/* Number of nodes in the computation */
int a0NodeCount;

/* Communicator used for the termination function */
static MPI_Comm _a0BarrierComm;

/* Tag used for the init function */
#define A0InitTag 1

/* Tag used for the termination function */
#define A0TerminationTag 2


/***************************************************************/
/* variables parsed in the command line, unused for the moment */
int _a0AccountLevel = 0;

/* whether athapascan should be verbose or not */
int _a0Verbose = 0;

/* verbose echoes the elapsed time. store the start time */
static struct timeval _a0StartTime;

/* whether athapascan should stop for debug attach or not */
static int _a0Stop = 0;

/* whether athapascan should divert std* or not */
static int _a0DivertStdIn = 0;
static int _a0DivertStdOut = 0;
static int _a0DivertStdErr = 0;

/* prefixes to diverted files */
static char *_a0DivertStdInPrefix = "stdin";
static char *_a0DivertStdOutPrefix = "stdout";
static char *_a0DivertStdErrPrefix = "stderr";

/* Abort service number. a0b calls this service to abort system */
static int _a0AbortService;

/* terminate mutex: only one thread can call a0Terminate() */
static a0tMutex _a0TerminateMutex;


/* Chain termination function. This function is called if any other
 * module of Athapascan-0 need a proper termination. See a0fmem.c, in
 * _a0UseDMA() function for an example.
 */
void (*_a0TerminateChain)(void) = NULL;


/****************************************************************************
		    Parses the Athapascan arguments
****************************************************************************/

static _a0ParseArguments(int *argcp, char ***argvp) {

int	argc, i, ii, ret;
char	**argv;
     
   /*
    * Parse command line looking for athapascan directives and
    * for akernel fonction mode. Values inside parenthesis are 
    * default:
    *    -aktimed  : set poll time for "timed daemon" (10000)
    *    -akdemand : turns off communication daemon (off)
    *    -akdaemon : turns on communication  daemon (on)
    *    -aktimeout: set time out for demand mode (no time-out equal to 0)
    */

     argc = *argcp;
     argv = *argvp;
    
     if ((ret=a0ArgFindName(argc, argv,"-a0verbose")) > 0) { 
        _a0Verbose=1;
        argv[ret]="used";
     }
     
     if ((ret=a0ArgFindName(argc, argv,"-a0stop")) > 0) {
        _a0Stop=1;
        argv[ret]="used";
     }
     
     if (ret=a0ArgGetInt(argc, argv, "-a0stack",(int *) &_a0DefaultStack)) {
        if (ret == -1)
	    _a0Error(A0ErrFatal, "args: Missing or bad stack size");
        if ((_a0DefaultStack < A0MinStackSize) ||
	    (_a0DefaultStack > A0MaxStackSize))
	    _a0Error(A0ErrFatal, "args: Invalid stack size");
	argv[ret]="used";argv[ret+1]="used";
     }
     
     if (ret = a0ArgGetInt(argc, argv, "-a0account", &_a0AccountLevel)) { 
        if ((ret == 1) || (_a0AccountLevel <=0 ))
	    _a0Error(A0ErrFatal, "args: Missing account level");
        argv[ret]="used";argv[ret+1]="used";
     }
       
     if ((ret=a0ArgFindName(argc, argv,"-a0divert")) > 0) {
        _a0DivertStdErr=1;
	_a0DivertStdOut=1;
	_a0DivertStdIn =1;
        argv[ret]="used";
     } 
     
     if ((ret=a0ArgFindName(argc, argv,"-a0divertstdin")) > 0) {
	_a0DivertStdIn =1;
        argv[ret]="used";
     }   
 
     if ((ret=a0ArgFindName(argc, argv,"-a0divertstdout")) > 0) {
	_a0DivertStdOut =1;
        argv[ret]="used";
     } 
            
     if ((ret=a0ArgFindName(argc, argv,"-a0divertstderr")) > 0) {
	_a0DivertStdErr =1;
        argv[ret]="used";
     }
     
     if ((ret=a0ArgGetStr(argc, argv, "-a0stdin",&_a0DivertStdInPrefix)) > 0){
        if (ret == -1)
	    _a0Error(A0ErrFatal, "args: Invalid prefix name");
        _a0DivertStdIn =1;
        argv[ret]="used";argv[ret+1]="used";
     }     
     
     if ((ret=a0ArgGetStr(argc, argv, "-a0stdout",&_a0DivertStdOutPrefix))> 0){
        if (ret == -1)
	    _a0Error(A0ErrFatal, "args: Invalid prefix name");
        _a0DivertStdOut =1;
        argv[ret]="used";argv[ret+1]="used";
     }
     
     if ((ret=a0ArgGetStr(argc, argv, "-a0stdout",&_a0DivertStdErrPrefix))> 0){
        if (ret == -1)
	    _a0Error(A0ErrFatal, "args: Invalid prefix name");
        _a0DivertStdErr =1;
        argv[ret]="used";argv[ret+1]="used";
     }    
                     	     	 
     if (ret = a0ArgGetInt(argc, argv, "-a0timeout", &akTimeOut)) {
        if ((ret == -1) || (akTimeOut <=0 ))
           _a0Error(A0ErrFatal, "args: Invalid time out");
	argv[ret]="used";argv[ret+1]="used"; 
     }	 
     
     akDaemonOn=1;
     if ((ret=a0ArgFindName(argc, argv,"-a0demand")) > 0) {
        akDaemonOn=0;
	argv[ret]="used"; 
     }
          
     if ((ret=a0ArgFindName(argc, argv,"-a0daemon")) > 0) {
        akDaemonOn=1;
	argv[ret]="used"; 
     }  
        
     if (ret = a0ArgGetInt(argc, argv, "-a0timed", &akTimePoll)) {
        if ((ret == -1) || (akTimePoll <=0 ))
           _a0Error(A0ErrFatal, "args: Invalid poll time");
	akTimed=1;
	argv[ret]="used";argv[ret+1]="used"; 
     }

     akAdjustTime=0;     
     if ((ret=a0ArgFindName(argc, argv,"-a0adjust")) > 0) {
        akAdjustTime=1;
	argv[ret]="used"; 
     }   
       
     akNoComm=0;  
     if ((ret=a0ArgFindName(argc, argv,"-a0nocomm")) > 0) {
        akNoComm=1;
	argv[ret]="used"; 
     }  
        	
     /* Pack argv vector getting of NULL values                  */	
	         
     for(i=0; i < *argcp; i++) {
        if (strcmp(argv[i],"used"))
	   continue;
	for(ii=i; ii < (*argcp-1); ii++)
	   argv[ii]=argv[ii+1]; 
	*argcp-=1;
	i--;
     }    
     
     return;
}


/************************************************************************/
/* This function implements a barrier, on a thread level.
 * Uses the communicator _a0BarrierComm.
 */

static a0tError _a0InternalBarrier(int tag)
{
	int node, rc, dummy;
	aktHandle h;

	a0Debug(("Start of barrier on node %d.\n", akNode));

	if (!akNode) {		/* Node 0 code */

		/* Receive the message "ok, I'm done" from all other nodes... */
		for (node = 1; node < akSize; node++) {
			a0Debug(("Node 0 receive from %d\n", node));
			rc = akRecv(&dummy, 1, MPI_INT, node, tag,
				    _a0BarrierComm, &h);
			if (rc)
				_a0Error(A0ErrKernelComm, "barrier1");
		}

		a0Debug(("Node 0 received from everyone\n"));

		/* ...and tell them "ok, you're really done now". */
		for (node = 1; node < akSize; node++) {
			a0Debug(("Node 0 send to %d\n", node));
			rc = akSend(&dummy, 1, MPI_INT, node, tag,
				    _a0BarrierComm, &h);
			if (rc)
				_a0Error(A0ErrKernelComm, "barrier2");
		}

		a0Debug(("Node 0 sent back to everyone\n"));

	} else {		/* All nodes except node 0 */

		/* Send to proc 0 the message "ok, I'm done"... */
		a0Debug(("Node %d send to 0\n", akNode));
		rc = akSend(&dummy, 1, MPI_INT, 0, tag,
			    _a0BarrierComm, &h);
		if (rc)
			_a0Error(A0ErrKernelComm, "barrier3");

		a0Debug(("Node %d sent to node 0\n", akNode));

		/* ...and wait for his answer "ok, you're really done now". */
		a0Debug(("Node %d receive from 0\n", akNode));
		rc = akRecv(&dummy, 1, MPI_INT, 0, tag,
			    _a0BarrierComm, &h);
		if (rc)
			_a0Error(A0ErrKernelComm, "barrier4");

		a0Debug(("Node %d received from node 0\n", akNode));

	}

	a0Debug(("Barrier finished on node %d.\n", akNode));

	return 0;
}


/************************************************************************/
/* Function to be called before ANY other Athapascan call.
 * The order in which the initialization functions are called is VERY important.
 * After return from _a0Init, the user can do his own initializations (like
 * creating formats, services etc...), and when he's done he calls _a0InitCommit
 * and after return he can start services and do the 'real' program.
 */

extern a0tError _a0Init(int *argcp, char ***argvp)
{
	a0tError rc;

	/* store start time */
	gettimeofday(&_a0StartTime, 0);
	
	/* take the Athapascan arguments out of the command line */
	_a0ParseArguments(argcp, argvp);
	
	/* Akernel init */
	akInit(argcp, argvp);

	/* Get node rank and number of nodes */
	a0SelfNode = akNode;
	a0NodeCount = akSize;

	if (akNoComm == 0) {
	   /* Create the communicator for the barrier in _a0Terminate
	    * (should be replaced by a port one of these days...)
	    */
	    akMPI_Comm_dup(MPI_COMM_WORLD, &_a0BarrierComm);
	
	   _a0InitServices();	   /* Services table init */
	}
	
	/* initialize thread specific data management */
	_a0InitSpecificData();	/* no traces before this call! */

	/* set up datatype table */
	_a0InitDatatypes();

	/* initialize prio & thread startup module */
	_a0InitPrio();
	_a0InitThread();
	
        if (akNoComm == 0) {
	
	   /* Init ports (communicators...) */
	   _a0InitPorts();
	
	   /* Your attention please: When non-unique ports are created, all processors
	    * must do the creation in the same order. So ports created in threads other
	    * than main may be a bad idea... :-)
	    */
	
	   /* SERV daemon init (daemon thread created and related port).  */
	   _a0InitServ();
	
	   /* URGENT daemon init (daemon thread created and related port).  */
	   _a0InitUrgent();
	 }
 
	 
	/*
	 * Input/output/error divert. The lines that follow divert
	 * the basic stdio files to disk files specified by the
	 * user. Each node will open a filename of the type
	 * "name-node", where name can be "stdin", "stdout",
	 * "stderr" or user-supplied strings and node is the node
	 * number.
	 */

	if (_a0DivertStdIn) {
		char name[40];
		int fd;
		sprintf(name, "%s-%d", _a0DivertStdInPrefix, akNode);
		fd = open(name, O_RDONLY, 0644);
		if (fd != (-1))
			dup2(fd, 0);
	}
	if (_a0DivertStdOut) {
		char name[40];
		int fd;
		sprintf(name, "%s-%d", _a0DivertStdOutPrefix, akNode);
		fd = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
		if (fd != (-1))
			dup2(fd, 1);
	}
	if (_a0DivertStdErr) {
		char name[40];
		int fd;
		sprintf(name, "%s-%d", _a0DivertStdErrPrefix, akNode);
		fd = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
		if (fd != (-1))
			dup2(fd, 2);
	}
/*
 * If Athapascan-0 is to be verbose, the every node echoes
 * a message with its pid an hostname. The node 0 echoes
 * the current version of the library.
 */

	if (_a0Verbose) {
		char line[80], name[40];

		if (!akNode) {
			sprintf(line, "Athapascan-0 %s (%s.%s) initiating\n",
				akVERSION, akSETUP_MPI, akSETUP_THREAD);
			fprintf(stderr, line);
		}
		gethostname(name, 40);
		sprintf(line, "ath0b node %d started on %s, pid %d\n",
			akNode, name, getpid());
		fprintf(stderr, line);
	}
	/*
	 * Debugging trick. The lines that follow stop the node 0
	 * until some data comes in a pipe named "a0stop". The
	 * command "a0cont" writes in that pipe, unblocking the
	 * node 0. All the nodes do a barrier, so no one runs until
	 * the node 0 unblocks.
	 */

	if (_a0Stop) {
		if (!akNode) {
			int pipe;
			char r;
			if (_a0Verbose)
				fprintf(stderr, "Athapascan-0 stopped. Run a0cont to continue\n");
			mknod("a0stop", S_IFIFO | 0600, 0);
			pipe = open("a0stop", O_RDONLY, 0);
			read(pipe, &r, 1);
			close(pipe);
			unlink("a0stop");
		}
		_a0InternalBarrier(A0InitTag);
	}
	

	/* create mutex for terminate */
	_a0NewMutex(&_a0TerminateMutex);

        if (akNoComm == 0) {
	   /* register the abort service */
	   _a0NewService(&_a0AbortService, (a0tRemoteFunction) _a0Terminate,
		         A0DefaultScheduling, A0DefaultPriority,
		         A0DefaultStack);
	}
	
	return 0;
}


/************************************************************************/
/* Termination of the user initialisations.
 * Upon return from this call, all the initializations are done on all the
 * nodes and everything is ready for the 'real' program to start.
 */

extern a0tError _a0InitCommit()
{
	/* The thread that called _a0InitCommit is the 'main' thread of the
	 * computation.
	 */

	/* Let's do a barrier so everyone is ready then */
	
	if (akNoComm == 0) 
	   _a0InternalBarrier(A0InitTag);

	return 0;
}


/************************************************************************/
/* This function must be called by all the 'main' threads of computation on
 * all the nodes of the computation.
 * NOTHING can be done in the 'main' thread after this call.
 * When all the nodes have called this function, the computation terminates,
 * all running threads are killed. It is the user responsibilty to assure that
 * this function is called (on the last node where it's called) AFTER all his
 * threads have completed.
 */

extern a0tError _a0Terminate(void)
{
	a0tError rc;

	

if (akNoComm == 0) {	

        a0Debug(("%d:a0Terminate before a0LockMutex\n", a0SelfNode));
	/* lock the mutex and unlock jamais */
	_a0LockMutex(&_a0TerminateMutex);

	/* We only do a thread-level barrier here, to wait for all the mains of
	 * all the nodes to terminate.
	 * We could turn off the daemons in some clean way, and have every node
	 * waiting for his daemons to turn off before existing.
	 * ...To be thought of...
	 */

        a0ThreadsTotal--;	/* main() is no longer active (user point of view) */

	a0Debug(("%d:Before Internal barrier\n", a0SelfNode));

	_a0InternalBarrier(A0TerminationTag);

	/* call the terminate chain if it was defined */
	if (_a0TerminateChain)
		_a0TerminateChain();

	/* terminate local daemons */
	
        a0Debug(("%d:Before Term Serv\n", a0SelfNode));

	_a0TermServ();
	a0Debug(("%d:After TermServ and Before TermUrg\n", a0SelfNode));
	
	_a0TermUrgent();
        a0Debug(("%d:After TermUrg\n", a0SelfNode));
	
#ifdef A0_TRACE
	/* Trace system */
	a0_perf_internal_terminate();
#endif				/* A0_TRACE */
}


	if (_a0Verbose) {
		struct timeval endtime;
		int hours, minutes;
		double seconds;
		char line[80];

		gettimeofday(&endtime, 0);
		seconds = (double) (endtime.tv_sec - _a0StartTime.tv_sec) +
		    0.000001 * (double) (endtime.tv_usec - _a0StartTime.tv_usec);
		hours = (int) (seconds / 3600);
		seconds -= (float) (hours * 3600);
		minutes = (int) (seconds / 60);
		seconds -= (float) (minutes * 60);

		sprintf(line, "ath0b node %d terminated, elapsed time %02d:%02d:%lf\n",
			akNode, hours, minutes, seconds);
		fprintf(stderr, line);
	}
	/* Turn off low level machinery */
	akFinalize();		/* This call may not return... */

	return 0;
}


/*************************************************************************
		_a0Abort - terminates abnormally Athapascan
*************************************************************************/

extern a0tError _a0Abort(int err)
{
	struct timeval endtime;
        int 		hours, minutes;
	double 		seconds;
	char 		line[80];

	gettimeofday(&endtime, 0);
	seconds = (double) (endtime.tv_sec - _a0StartTime.tv_sec) +
	    0.000001 * (double) (endtime.tv_usec - _a0StartTime.tv_usec);
	hours = (int) (seconds / 3600);
	seconds -= (float) (hours * 3600);
	minutes = (int) (seconds / 60);
	seconds -= (float) (minutes * 60);

	sprintf(line, "ath0b node %d ABORTING, elapsed time %02d:%02d:%lf\n",
		akNode, hours, minutes, seconds);
	fprintf(stderr, line);
	
	akMPI_Abort(MPI_COMM_WORLD, 0);
}

