/*
 *
 * a0prio.c : priority control functions
 *
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 21, 1996
 *
 */

#include <akernel.h>
#include <errno.h>

#define A0_SOURCE
#include "a0error.h"
#include "a0prio.h"
#include "a0thread.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)


/* Default values (Athapascan wide) for thread and service creation */
a0tScheduling	_a0DefaultScheduling;
int		_a0DefaultPriority;
size_t		_a0DefaultStack = 0;


/* Define the initial pthread priority of each policy and
   the max user priorities of each policy */

int A0MaxFifoPriority = 0;		/* FIFO */
static int _a0PriFifo0;

int A0MaxRRPriority = 0;		/* Round Robin */
static int _a0PriRR0;

int A0MaxOtherPriority = 0;		/* Other (Default) */
static int _a0PriOther0;


/* Priorities reserved for Athapascan.
 * It is possible that these priorities be used by Athapascan programs
 * (depending on the values of _a0Pri???Max and _a0Pri???0), if the pthread
 * package used does not provide enough different priorities.
 * If pthreads provides enough priorities, Athapascan user priorities will
 * all be lower than _a0PriDaemonLow, and higher than the _a0PriLowest.
 */
int _a0PriDaemonHigh, _a0PriDaemonMed, _a0PriDaemonLow;
int _a0PriLowest;

/* the scheduling to be used with those priorities */
a0tScheduling _a0DaemonScheduling;



#define A0Max(x, y) ((x)>(y) ? (x) : (y))
#define A0Min(x, y) ((x)<(y) ? (x) : (y))

  
/************************************************************************
	_a0SetMinMaxPrio - adjust valid range of priorities
************************************************************************/

static a0tError _a0SetMinMaxPrio() {

  int max, min;

  /* Get pthread's max and min priorities, and set Athapascan max and
   * min accordingly, after reserving some for Athapascan use.
   */

#ifdef akSCHED_FIFO
  /* Athapascan kernel uses akSCHED_PRI_FIFO_MAX_NP */
  _a0PriDaemonHigh = A0Max(akSCHED_PRI_FIFO_MAX_NP - 1, akSCHED_PRI_FIFO_MIN_NP);
  _a0PriDaemonMed  = A0Max(_a0PriDaemonHigh - 1, akSCHED_PRI_FIFO_MIN_NP);
  _a0PriDaemonLow  = A0Max(_a0PriDaemonMed - 1, akSCHED_PRI_FIFO_MIN_NP);
  _a0PriLowest     = akSCHED_PRI_FIFO_MIN_NP;
  _a0DaemonScheduling   = akSCHED_FIFO;
  max		   = A0Max(_a0PriDaemonLow - 1, akSCHED_PRI_FIFO_MIN_NP);
  min		   = A0Min(akSCHED_PRI_FIFO_MIN_NP + 1, max);
  A0MaxFifoPriority	= max - min;
  _a0PriFifo0		= min;

  a0Debug(("akSCHED_PRI_FIFO_MIN_NP=%d akSCHED_PRI_FIFO_MAX_NP=%d\n",
		akSCHED_PRI_FIFO_MIN_NP, akSCHED_PRI_FIFO_MAX_NP));
  a0Debug(("_a0PriFifo0=%d A0MaxFifoPriority=%d\n",
		_a0PriFifo0, A0MaxFifoPriority));

#endif /* akSCHED_FIFO */


#ifdef akSCHED_RR
#ifndef akSCHED_FIFO
  /* Athapascan kernel uses akSCHED_PRI_RR_MAX_NP */
  _a0PriDaemonHigh = A0Max(akSCHED_PRI_RR_MAX_NP - 1, akSCHED_PRI_RR_MIN_NP);
  _a0PriDaemonMed  = A0Max(_a0PriDaemonHigh - 1, akSCHED_PRI_RR_MIN_NP);
  _a0PriDaemonLow  = A0Max(_a0PriDaemonMed - 1, akSCHED_PRI_RR_MIN_NP);
  _a0PriLowest     = akSCHED_PRI_RR_MIN_NP;
  max		   = A0Max(_a0PriDaemonLow - 1, akSCHED_PRI_RR_MIN_NP);
  min		   = A0Min(akSCHED_PRI_RR_MIN_NP + 1, max);
  _a0DaemonScheduling   = akSCHED_RR;
#else
  max		   = akSCHED_PRI_RR_MAX_NP;
  min		   = akSCHED_PRI_RR_MIN_NP;
#endif
  A0MaxRRPriority	= max - min;
  _a0PriRR0		= min;

  a0Debug(("akSCHED_PRI_RR_MIN_NP=%d akSCHED_PRI_RR_MAX_NP=%d\n",
		akSCHED_PRI_RR_MIN_NP, akSCHED_PRI_RR_MAX_NP));
  a0Debug(("_a0PriRR0=%d A0MaxRRPriority=%d\n",
		_a0PriRR0, A0MaxRRPriority));

#endif /* akSCHED_RR */


#ifdef akSCHED_OTHER
#if !defined(akSCHED_FIFO) && !defined(akSCHED_RR)
  /* Athapascan kernel uses akSCHED_PRI_OTHER_MAX_NP */
  _a0PriDaemonHigh = A0Max(akSCHED_PRI_OTHER_MAX_NP - 1, akSCHED_PRI_OTHER_MIN_NP);
  _a0PriDaemonMed  = A0Max(_a0PriDaemonHigh - 1, akSCHED_PRI_OTHER_MIN_NP);
  _a0PriDaemonLow  = A0Max(_a0PriDaemonMed - 1, akSCHED_PRI_OTHER_MIN_NP);
  _a0PriLowest     = akSCHED_PRI_OTHER_MIN_NP;
  _a0DaemonScheduling   = akSCHED_OTHER;
  max		   = A0Max(_a0PriDaemonLow - 1, akSCHED_PRI_OTHER_MIN_NP);
  min		   = A0Min(akSCHED_PRI_OTHER_MIN_NP + 1, max);
#else
  max		   = akSCHED_PRI_OTHER_MAX_NP;
  min		   = akSCHED_PRI_OTHER_MIN_NP;
#endif
  A0MaxOtherPriority	= max - min;
  _a0PriOther0		= min;

  a0Debug(("akSCHED_PRI_OTHER_MIN_NP=%d akSCHED_PRI_OTHER_MAX_NP=%d\n",
		akSCHED_PRI_OTHER_MIN_NP, akSCHED_PRI_OTHER_MAX_NP));
  a0Debug(("_a0PriOther0=%d A0MaxOtherPriority=%d\n",
		_a0PriOther0, A0MaxOtherPriority));

#endif /* akSCHED_OTHER */

#if !defined(akSCHED_FIFO) && !defined(akSCHED_RR) && !defined(akSCHED_OTHER)
  error! at least one policy must be defined.
#endif

  a0Debug(("_a0PriDaemonHigh=%d _a0PriDaemonMed=%d _a0PriDaemonLow=%d\n"
		"_a0PriLowest=%d _a0DaemonScheduling=%d\n",
  		_a0PriDaemonHigh, _a0PriDaemonMed,
  		_a0PriDaemonLow, _a0PriLowest,
  		_a0DaemonScheduling));

  return A0ErrOk;
}


/*************************************************************************
	_a0InitPrio - Init prio module
*************************************************************************/

extern a0tError _a0InitPrio(void) {

  a0tError ret;
  int prio, sched;
  struct aksched_param params;

  /* Set min and max priorities for all scheduling modes */
  _a0SetMinMaxPrio();

  /* The defaults initialization should take place before the parse
   * of the command line, so the command line can modify it.
   * The test of the variable being ==0 or not (like is done for the stack
   * in the following lines...) is not applicable to _a0DefaultScheduling and
   * _a0DefaultPriority, so the 'default default' should be set before the
   * 'command line default'. IG.
   */

  /* if no default stack size, find one. */
  if (!_a0DefaultStack) {
    _a0DefaultStack = akPTHREAD_STACK_DEFAULT_NP;
    a0Debug(("_a0InitPrio: default stack %d\n", _a0DefaultStack));
  }

  /* get user prio & sched in terms os a0b and 
     set default Scheduling and priority */
  akpthread_getschedparam(akpthread_self(), &sched, &params);
  if (_a0Pthread2User(sched, params.aksched_priority,
		&_a0DefaultScheduling, &_a0DefaultPriority))
	_a0Error(A0ErrThreadAttr, "main prio");

  a0Debug(("_a0InitPrio: main thread (default) prio=%d sched=%d\n",
	     _a0DefaultPriority, _a0DefaultScheduling));

  return A0ErrOk;
}


/*****************************************************************************
_a0User2Pthread - Converts user priority&policy values to pthread
*****************************************************************************/

extern a0tError _a0User2Pthread(a0tScheduling sched, int UserPrio,
				int *PthreadSched, int *PthreadPrio) {

/* The scheduling is necessary because the conversion is based on it.
 * An error is returned if the requested (user) priority is out of range.
 */


  if (UserPrio != A0InheritPriority) {
    if (UserPrio<0) return A0ErrInvPriority;

    switch (sched) {
      case A0DaemonScheduling:
	*PthreadPrio = UserPrio;
	*PthreadSched = _a0DaemonScheduling;
	return A0ErrOk;
#ifdef akSCHED_FIFO
      case A0FifoScheduling:
	if (UserPrio>A0MaxFifoPriority) return A0ErrInvPriority;
	*PthreadPrio = UserPrio + _a0PriFifo0;
	*PthreadSched = sched;
	break;
#endif
#ifdef akSCHED_RR
      case A0RRScheduling:
	if (UserPrio>A0MaxRRPriority) return A0ErrInvPriority;
	*PthreadPrio = UserPrio + _a0PriRR0;
	*PthreadSched = sched;
	break;
#endif
#ifdef akSCHED_OTHER
      case A0OtherScheduling:
	if (UserPrio>A0MaxOtherPriority) return A0ErrInvPriority;
	*PthreadPrio = UserPrio + _a0PriOther0;
	*PthreadSched = sched;
	break;
#endif
      case A0InheritScheduling:
      	/* We don't test bound here... it's bad, but that's life. */
	/* Even worse... I'm doing it _very_ dirty: if you inherit scheduling,
	 * you inherit priority. Horrible, isn't it?
	 */
	/* What should be done here is to test what is the priority we're
	 * inheriting, and depending upon it, compute the correct value
	 * of *PthreadPrio, function of UserPrio.
	 * If UserPrio == A0InheritPriority, then *PthreadPrio should take
	 * A0InheritPriority, but in the other cases it shouldn't.
	 */
	{
	  struct aksched_param params;
	  int                  schedul;

	  akpthread_getschedparam(akpthread_self(), &schedul, &params);
	  *PthreadPrio = params.aksched_priority;
	  *PthreadSched = schedul;
	}
	break;
      default:
    	return A0ErrInvPriority;
    }
  }
  else { /* UserPrio == A0InheritPriority */
    *PthreadPrio = A0InheritPriority; /* It'll be taken care of later */
    *PthreadSched = sched;
  }

  return A0ErrOk;
}


/*****************************************************************************
_a0Pthread2User - Converts pthread priority&policy values to user
*****************************************************************************/

extern  a0tError _a0Pthread2User(int PthreadSched, int PthreadPrio,
				a0tScheduling *UserSched,
				int *UserPrio) {

/* The scheduling is necessary because the conversion is based on it.
 */

  switch (PthreadSched) {
#ifdef akSCHED_FIFO
  case A0FifoScheduling:
    *UserPrio = PthreadPrio - _a0PriFifo0;
    break;
#endif
#ifdef akSCHED_RR
  case A0RRScheduling:
    *UserPrio = PthreadPrio - _a0PriRR0;
    break;
#endif
#ifdef akSCHED_OTHER
  case A0OtherScheduling:
    *UserPrio = PthreadPrio - _a0PriOther0;
      break;
#endif
  default:
    return A0ErrInvPriority;
  }

  *UserSched = PthreadSched;
  return A0ErrOk;
}


/******************************************************************************
	_a0IllegalPriority - Checks if prio is valid for given sched mode
******************************************************************************/

extern a0tError _a0IllegalPriority(a0tScheduling sched, int prio) {

  /* Checks that for the given sched mode, prio has a valid value.
   * This function should be called ONLY for user threads, not for Athapascan
   * threads, that have priorities outside of user allowed range.
   */

  a0Debug(("Entering _a0IllegalPriority. sched=%d, prio=%d\n", sched, prio));

  if (prio<0) return A0ErrInvPriority;

  switch (sched) {
#ifdef akSCHED_FIFO
    case A0FifoScheduling:
	if (prio>A0MaxFifoPriority) return A0ErrInvPriority;
	break;
#endif
#ifdef akSCHED_RR
    case A0RRScheduling:
	if (prio>A0MaxRRPriority) return A0ErrInvPriority;
	break;
#endif
#ifdef akSCHED_OTHER
    case A0OtherScheduling:
	if (prio>A0MaxOtherPriority) return A0ErrInvPriority;
	break;
#endif
    default:
	return A0ErrInvPriority;
  }

  return A0ErrOk;
}


/*************************************************************************
	_a0GetThreadPriority - returns the priority of a thread
*************************************************************************/

extern a0tError _a0GetThreadPriority(a0tThread *thread, int *priop) {

  int psched;
  a0tScheduling sched;
  struct aksched_param params;

  akpthread_getschedparam(thread->pth_id, &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, &sched, priop);
  return A0ErrOk;
}


/*************************************************************************
	_a0GetThreadScheduling - returns the scheduling of a thread
*************************************************************************/

extern a0tError _a0GetThreadScheduling(a0tThread *thread,
					a0tScheduling *schedp) {

  int prio, psched;
  a0tScheduling sched;
  struct aksched_param params;

  akpthread_getschedparam(thread->pth_id, &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, schedp, &prio);
  return A0ErrOk;
}


/*************************************************************************
	_a0GetSelfThreadPriority - returns the priority of a thread
*************************************************************************/

extern a0tError _a0GetSelfThreadPriority(int *priop) {

  int psched;
  a0tScheduling sched;
  struct aksched_param params;

  akpthread_getschedparam(akpthread_self(), &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, &sched, priop);
  return A0ErrOk;
}


/*************************************************************************
	_a0GetSelfThreadScheduling - returns the scheduling of a thread
*************************************************************************/

extern a0tError _a0GetSelfThreadScheduling(a0tScheduling *schedp) {

  int prio, psched;
  a0tScheduling sched;
  struct aksched_param params;

  akpthread_getschedparam(akpthread_self(), &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, schedp, &prio);
  return A0ErrOk;
}


/*************************************************************************
	_a0SetThreadPriority - sets the priority of a thread
*************************************************************************/

extern a0tError _a0SetThreadPriority(a0tThread *thread, int prio) {

  int pprio, psched, oldprio;
  a0tScheduling sched;
  a0tError r;
  struct aksched_param params;

  akpthread_getschedparam(thread->pth_id, &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, &sched, &oldprio);
  r = _a0User2Pthread(sched, prio, &psched, &pprio);
  if (r) return r;

  params.aksched_priority = pprio;
  akpthread_setschedparam(thread->pth_id, psched, &params);
  return A0ErrOk;
}


/*************************************************************************
	_a0SetThreadScheduling - sets the scheduling of a thread
*************************************************************************/

extern a0tError _a0SetThreadScheduling(a0tThread *thread,
					a0tScheduling sched,
					int prio) {

  int pprio, psched;
  a0tError r;
  struct aksched_param params;

  r = _a0User2Pthread(sched, prio, &psched, &pprio);
  if (r) return r;

  params.aksched_priority = pprio;
  akpthread_setschedparam(thread->pth_id, psched, &params);
  return A0ErrOk;
}


/*************************************************************************
	_a0SetSelfThreadPriority - sets the priority of a thread
*************************************************************************/

extern a0tError _a0SetSelfThreadPriority(int prio) {

  int pprio, psched, oldprio;
  a0tScheduling sched;
  a0tError r;
  struct aksched_param params;

  akpthread_getschedparam(akpthread_self(), &psched, &params);
  _a0Pthread2User(psched, params.aksched_priority, &sched, &oldprio);
  r = _a0User2Pthread(sched, prio, &psched, &pprio);
  if (r) return r;

  params.aksched_priority = pprio;
  akpthread_setschedparam(akpthread_self(), psched, &params);
  return A0ErrOk;
}


/*************************************************************************
	_a0SetSelfThreadScheduling - sets the scheduling of a thread
*************************************************************************/

extern a0tError _a0SetSelfThreadScheduling(a0tScheduling sched,
						int prio) {

  int pprio, psched;
  a0tError r;
  struct aksched_param params;

  r = _a0User2Pthread(sched, prio, &psched, &pprio);
  if (r) return r;

  params.aksched_priority = pprio;
  akpthread_setschedparam(akpthread_self(), psched, &params);
  return A0ErrOk;
}
