/*
 *
 * a0semap.c : semaphore & mutex starting functions
 *
 *
 * by Marcelo Pasin
 * Apache Project, IMAG, Grenoble, France
 * Feb 21, 1996
 *
 */

#include <akernel.h>
#include <errno.h>

#define A0_SOURCE
#include "a0error.h"
#include "a0semap.h"
#include "a0thread.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)


/*************************************************************************
	_a0NewSemaphore - creates a new semaphore
*************************************************************************/

extern a0tError _a0NewSemaphore(a0tSemaphore *semap, int value) {

  /* any errors here are fatal */
  if (akSemaphoreCreate(semap, value))
	_a0Error(A0ErrSemapCreate, "thread");

  return A0ErrOk;
}


/*************************************************************************
	_a0DisposeSemaphore - disposes a semaphore
*************************************************************************/

extern a0tError _a0DisposeSemaphore(a0tSemaphore *semap) {

  /* any errors here are fatal */
  if (akSemaphoreDestroy(semap))
	_a0Error(A0ErrSemapDestr, "thread");

  return A0ErrOk;
}


/*************************************************************************
	_a0PSemaphore - gets a semaphore (do P)
*************************************************************************/

extern a0tError _a0PSemaphore(a0tSemaphore *semap) {

  a0ThreadsBlockedLock++;
  /* any errors here are fatal */
  if (akSemaphoreP(semap))
	_a0Error(A0ErrSemapP, "thread");
  a0ThreadsBlockedLock--;

  return A0ErrOk;
}


/*************************************************************************
	_a0TryPSemaphore - tries to get a semaphore (try P)
*************************************************************************/

extern a0tError _a0TryPSemaphore(a0tSemaphore *semap, int *success) {

  /* any errors here are fatal */
  if (akSemaphoreTryP(semap, success))
	_a0Error(A0ErrSemapP, "try thread");

  return A0ErrOk;
}


/*************************************************************************
	_a0VSemaphore - gives a semaphore (do V)
*************************************************************************/

extern a0tError _a0VSemaphore(a0tSemaphore *semap) {

  /* any errors here are fatal */
  if (akSemaphoreV(semap))
	_a0Error(A0ErrSemapV, "thread");

  return A0ErrOk;
}


/*************************************************************************
	_a0NewMutex - creates a new mutex
*************************************************************************/

extern a0tError _a0NewMutex(a0tMutex *mutex) {

  if (akpthread_mutex_init(mutex, akPTHREAD_MUTEXATTR_DEFAULT_NP))
	return A0ErrNoMem;

  return A0ErrOk;
}


/*************************************************************************
	_a0DisposeMutex - disposes a mutex
*************************************************************************/

extern a0tError _a0DisposeMutex(a0tMutex *mutex) {

  int r;

  if (r = akpthread_mutex_destroy(mutex))
	return (r==EINVAL) ? A0ErrInvalid : A0ErrLockedMutex;
  return A0ErrOk;
}


/*************************************************************************
	_a0LockMutex - locks a mutex
*************************************************************************/

extern a0tError _a0LockMutex(a0tMutex *mutex) {

  int ret;

  a0ThreadsBlockedLock++;
  if (ret = akpthread_mutex_lock(mutex))
	return (ret==EINVAL) ? A0ErrInvalid : A0ErrLockedMutex;
  a0ThreadsBlockedLock--;

  return A0ErrOk;
}


/*************************************************************************
	_a0UnlockMutex - unlocks a mutex
*************************************************************************/

extern a0tError _a0UnlockMutex(a0tMutex *mutex) {

  if (akpthread_mutex_unlock(mutex)) 
	return A0ErrInvalid;

  return A0ErrOk;
}


/*************************************************************************
	_a0ContextDestructor - destroys a context variable
*************************************************************************/

void _a0ContextDestructor(void *c) {
  /* this function should destroy the context variable associated with
     a thread when it terminates. There are two solutions: a0b can do
     nothing or it can ask this function in the user interface of
     _a0NewContext. I decided to don't ask. */
}


/*************************************************************************
	_a0NewContext - creates a context descriptor
*************************************************************************/

extern a0tError _a0NewContext(a0tContext *contextp) {

  /* This function creates a context descriptor. Each thread can
     associate a private context variable to a context descriptor. This
     allows to associate a per-thread thread-private data structure.
  */
  if (akpthread_key_create(contextp, _a0ContextDestructor))
    return A0ErrNoMem;

  return A0ErrOk;
}


/*************************************************************************
	_a0SetSelfContext - associates data to context
*************************************************************************/

extern a0tError _a0SetSelfContext(a0tContext *contextp, void *datap) {

  /* This function associates a thread private data structure to a
     given context descriptor. Each thread can associate its own data
     stucture to a context descriptor. */
 
  switch (akpthread_setspecific(*contextp, datap)) {
  case ENOMEM:
    return A0ErrNoMem;
  case EINVAL:
    return A0ErrInvalid;
  default:
    return A0ErrOk;
  }
}


/*************************************************************************
	_a0GetSelfContext - gets data associated to context
*************************************************************************/

extern a0tError _a0GetSelfContext(a0tContext *contextp, void **datapp) {

  /* This function returns a thread private data structure which is
     associated to a given context descriptor. */
 
  if (akpthread_getspecific_np(*contextp, datapp))
    return A0ErrInvalid;
  return A0ErrOk;
}
