/*
 *
 * a0specific.h : Include for _a0specific.c
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 17, 1996
 *
 */


/* Eliminates multiple includes of this file */
#ifndef _A0SPECIFIC_H
#define _A0SPECIFIC_H

#include <akernel.h>

#ifdef A0_REUSETHR
#include <setjmp.h>
#endif

#include "a0services.h"
#include "a0error.h"
#include "a0thread.h"
#include "a0semap.h"
#include "a0buffer.h"


/* Thread specific data structure.
 * All threads in a computation except main threads (plural because there's
 * one main thread on every node) have specific data.
 */
typedef struct _a0tThreadSpecific {

  /* Next free entry if structure is on free list */
  struct _a0tThreadSpecific *nextfree;

  int			flags;	/* See below */
  a0tBuffer		buffer; /* Valid with flag A0SpecificBuffer */
  void			*param;	/* User parameter passed to the thread */
  a0tFunction		function; /* Function the thread executes */
  void			*result; /* result of the exit function */  
#ifdef A0_TRACE
  unsigned long		logical_id; /* Logical thread id */
#endif
  a0tMutex		detach;	/* mutex to detach & exit */
#ifdef A0_REUSETHR
  int			Prio;	/* Thread Priority */
  a0tSemaphore		join;	/* semaphore to wait on join */
  a0tSemaphore		reuse;	/* semaphore to block and reuse */
  jmp_buf		begin;  /* begin of reuse loop */
#endif

} a0tThreadSpecific;


#ifdef A0_SOURCE
/***************************************************************************
 * Defines only used for compiling Athapascan0
 **************************************************************************/


/* Max number of service threads active on any single node at any time.
 * This value is needed because we maintain an array with the thread-
 * specific data for the service threads.
 * One day the handling of this array could (should?) be rewritten to
 * allow for dynamic growth. But this day hasn't come yet (or this comment
 * would have been deleted...).
 * Another option could be to store these data in the thread's stack. But
 * this implies we know the address of the stack before the service thread
 * starts, and we can use a few bytes of this space for the thread specific
 * data, and then make the thread beleive the stack is a bit smaller.
 * Yet another option is that when the service thread starts it copies the
 * data directly form the daemon's message buffer (or wherever the data are)
 * and then signals the daemon the copy is done (the daemon will wait for this
 * signal before continuing). This was not used because it involves too many
 * context switches and when the service thread is not scheduled immediately
 * it can lock the daemon.
 */
#define A0MAXSERVICETHREADS	1000


/**************************************************************************/
/* Thread specific structure flags */

/* wether the structure is used or not */
#define A0SpecificUsed 1

/* if the structure has a buffer pointer or not */
#define A0SpecificBuffer 2

/* whether the thread is detached */
#define A0SpecificDetached 4

/* whether the thread has exited or not */
#define A0SpecificExited 8


/**************************************************************************
 * Type definitions
 **************************************************************************/

/**************************************************************************
 * Variables definitions
 **************************************************************************/

/* Array containing the thread specific data */
extern a0tThreadSpecific _a0Specific[];

/* Head of free list of thread specific data */
extern a0tThreadSpecific *_a0SpecificFreeHead;

/* Mutex protecting _a0SpecificFreeHead (and the rest of the list...) */
extern akpthread_mutex_t _a0SpecificMutex;

/* Key value for per-thread context */
extern akpthread_key_t _a0threadSpecificKey;

/************************************************************************
 * Functions definitions
 ************************************************************************/

/* Initialise the thread specific structure (to be called once) */
extern a0tError _a0InitSpecificData();

/* Copy a specific data structure */
extern a0tError _a0CopySpecificData(a0tThreadSpecific *dest,
				   a0tThreadSpecific *source);

/* Allocate a thread specific data structure. */
extern a0tError _a0NewSpecificData(a0tThreadSpecific **data);

/* Free the pointed to specific thread data */
extern a0tError _a0FreeSpecificData(a0tThreadSpecific *data);

/* Get a pointer on the thread specific structure for the calling thread. */
extern a0tError _a0GetSpecificData(a0tThreadSpecific **data);

/* Set a pointer on the thread specific structure for the calling thread. */
extern a0tError _a0SetSpecificData(a0tThreadSpecific *data);


#ifdef A0_REUSETHR
/* Allocate a thread specific (with prio) data structure. */
extern a0tError _a0NewSpecificPrio(a0tThreadSpecific **data, int prio);

/* Free the pointed to specific (with prio) thread data */
extern a0tError _a0FreeSpecificPrio(a0tThreadSpecific *data, int prio);
#endif /* A0_REUSETHR */


#endif /* A0_SOURCE */


#endif /* _A0SPECIFIC_H */
