/*
 *
 * a0thread.c : thread handling functions
 *
 *
 * by Marcelo Pasin and Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Feb 21, 1996
 *
 */

#include <akernel.h>
#include <errno.h>

#ifdef A0_REUSETHR
#include <setjmp.h>
#endif

#define A0_SOURCE
#include "a0error.h"
#include "a0thread.h"
#include "a0prio.h"
#include "a0specific.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a) 


/*************************************************************************
 * Informational variables:
 * Counters that give the number of threads in various positions...
 * (counting only USER threads. Athapascan daemons not included).
 * These values may be completely irrelevant to the actual number of threads,
 * because these variables are updated WITHOUT any locking mecanism.
 */

/* total number of user threads in the node */
int a0ThreadsTotal;
/* number of threads blocked on a communication */
int a0ThreadsBlockedComm;
/* number of threads blocked on a semaphore, mutex or condition variable */
int a0ThreadsBlockedLock;


  
/*************************************************************************
	_a0InitThread - Init thread module
*************************************************************************/

a0tError _a0InitThread(void) {

  /* Init counting variables for the number of threads in different
   * categories */
  a0ThreadsTotal = 1; /* The main() thread executing this code */
  a0ThreadsBlockedComm = 0;
  a0ThreadsBlockedLock = 0;
  /* It's maybe one of the last times in the execution of the program that the
   * values in the above 3 variables really reflect the status of the system :-)
   */

  return A0ErrOk;
}


/*************************************************************************
	_a0NewThreadAttr - build attributes to create a thread
*************************************************************************/

static a0tError _a0NewThreadAttr(a0tThreadAttr *attr, a0tScheduling sched,
				int prio, size_t stack) {

/* The parameters can be a valid value for the parameter or A0InheritXxx,
 * in which case the value of the calling thread for that parameter is used.
 * (The use of A0InheritXxx is legal only when creating slaves or daemons.
 * for service invocation these values are not permitted).
 */

  struct aksched_param params;
  int newsch;

  /* create an attribute */
  if (akpthread_attr_init(attr))
	_a0Error(A0ErrThreadAttr, "create");

  a0Debug(("_a0NewThreadAttr sched %d prio %d stack %d\n",
		sched, prio, stack));

  /* test if should not inherit parent sched & prio */
  if ((sched != A0InheritScheduling) || (prio != A0InheritPriority)) {
    if (akpthread_attr_setinheritsched(attr, akPTHREAD_EXPLICIT_SCHED))
	_a0Error(A0ErrThreadAttr, "expl");
    if (sched == A0InheritScheduling) {
      if (akpthread_getschedparam(akpthread_self(), &newsch, &params))
	_a0Error(A0ErrThreadAttr, "getprio");
      sched = newsch;
    }
    if (prio == A0InheritPriority) {
      if (akpthread_getschedparam(akpthread_self(), &newsch, &params))
	_a0Error(A0ErrThreadAttr, "getprio");
      prio = params.aksched_priority;
    }

    /* set attributes */
    if (akpthread_attr_setschedpolicy(attr, sched))
	_a0Error(A0ErrThreadAttr, "sched");

#if defined(A0_TSOL2) || defined(A0_TLIN2) || \
    defined(A0_TSGI6) || defined(A0_TAIX4)	
       akpthread_attr_setscope(attr, akPTHREAD_SCOPE_SYSTEM);
#else
       akpthread_attr_setscope(attr, akPTHREAD_SCOPE_PROCESS);
#endif

	
#if !defined(A0_TAIX4) && !defined(A0_TLIN2) \
    && !defined(A0_TSGI6) && !defined(A0_TSOL2) 
    params.aksched_priority = prio;
    if (akpthread_attr_setschedparam(attr, &params))
	_a0Error(A0ErrThreadAttr, "prio");
#endif
  }
  else {
    if (akpthread_attr_setinheritsched(attr, akPTHREAD_INHERIT_SCHED))
	_a0Error(A0ErrThreadAttr, "inher");
  }

  /* Stack may be inherited too. So we set the stack part of the attribute
   * only if the stack should not be inherited.
   */

  if (stack != A0InheritStack)
    if (akpthread_attr_setstacksize(attr, stack))
	_a0Error(A0ErrThreadAttr, "stack");

#ifdef A0_TAIX4
  if (akpthread_attr_setdetachstate(attr, akPTHREAD_CREATE_JOINABLE))
	_a0Error(A0ErrThreadAttr, "detach");
#endif


  return A0ErrOk;
}


/*************************************************************************
	_a0ThreadLauncher - launcher function for slave/service threads
*************************************************************************/

static int _a0ThreadLauncher(void *arg) {

/* This function extracts from the thread specific data the adress of the
 * slave to call, the parameters to pass it, and does the call. It also
 * takes care of the cleanup of thread specific structure at termination
 * of the thread.
 */

  /* We type cast the argument into something usable */
  a0tThreadSpecific *specific_data = (a0tThreadSpecific *) arg;

  /* We set the thread specific context */
  _a0SetSpecificData(specific_data);

#ifdef A0_REUSETHR
begin_setjump:
  a0Debug(("_a0ThreadLauncher: before setjump.\n"));

  if (setjmp(specific_data->begin)) {
    /* it's coming from the longjmp */

    a0Debug(("_a0ThreadLauncher: hey! I'm saving a thread! (%08X)\n",
		specific_data));

    goto begin_setjump;
  }
#endif

#ifdef A0_TRACE
  akTraceSetThreadId(specific_data->logical_id);
#endif /* A0_TRACE */

  a0Debug(("_a0ThreadLauncher: Starting thread (%08X)\n", specific_data));

  /* We call the 'real' thread function, passing it the 'real' parameter */

  arg = (void *) specific_data->function(specific_data->param);

  a0Debug(("_a0ThreadLauncher: after call to slave\n"));

  /* Execute exit even when the thread terminates normally */
  _a0ExitThread(arg);
}


/*************************************************************************
	_a0NewSlave - starts a new thread with specific USER values
*************************************************************************/

extern a0tError _a0NewSlave(a0tThread *thread,
			int sched, int prio, size_t stack,
			a0tFunction function, void *value) {


/* This is the function that must be used by Athapascan programs, as the
 * priority value passed here is an Athapascan priority, and not a posix
 * priority.
 */

  a0tThreadSpecific *thread_data;
  a0tError ret;

  /* Create specific data for the new thread */

  ret = _a0NewSpecificData(&thread_data);
  if (ret) return ret;
  thread_data->param = value; /* Parameter passed to the created thread */
  thread_data->function = function; /* Code executed by slave. */
  thread_data->flags &= ~A0SpecificBuffer; /* No buffer for a slave thread */
  ret = _a0StartThreadPlease(thread, sched, prio, stack, thread_data);

  if (ret) {
    _a0FreeSpecificData(thread_data);
  }
  return ret;
}


/*************************************************************************
	_a0StartThreadPlease - start or reuse a thread
*************************************************************************/

extern a0tError _a0StartThreadPlease(a0tThread *thread,
				int sched, int prio, int stack,
				void *data) {

  /* data must be (void *): a0thread.h doesn't know a0tThreadSpecific */
  a0tThreadSpecific *thread_data = data;
  a0tError ret;
  int pth_pri, pth_sch;
#ifdef A0_REUSETHR
  a0tThreadSpecific *new_spec;
#endif

  /* This function is called by _a0NewSlave and by the Services daemon,
  in order to launch a new thread. The thread has already a Specific
  allocated. Threads are reused if possible. The reusal is done in a
  kinda dirty way: every time a new thread must be created, a new
  ThreadSpecific structure must be allocated via _a0NewSpecificData.
  Then it's filled in with the function to call and its parameter. This
  function then is called (_a0StartThreadPlease()). If the ThreadSpecific
  structure was already used before, a thread is blocked in a
  semaphore, waiting the structure to be filled to restart running. If
  the structure is shiny new, a new thread must be created. At the
  _a0ExitThread() call will block the thread and free the ThreadSpecific
  structure for future reuse. */

  /* Convert the 'user' priority into a pthread priority... */
  
#if !defined(A0_TSOL2) && !defined(A0_TLIN2) \
    && !defined(A0_TSGI6) && !defined(A0_TAIX4) 
  if (ret = _a0User2Pthread(sched, prio, &pth_sch, &pth_pri))
	return ret;
#endif


#ifdef A0_REUSETHR

  /* A specific structure must be reused. Try to get an used one. */
  if (!_a0NewSpecificPrio(&new_spec, prio)) {
    a0Debug(("Found a reusable specific %X for %d\n", new_spec, prio));

    /* found an used structure, copy old to it */
    new_spec->flags = thread_data->flags | A0SpecificUsed;
    new_spec->function = thread_data->function;

    /* if there is a buffer, its address mus be passed in parameters */
    if (new_spec->flags & A0SpecificBuffer) {
      /* I hope a buffer descriptor can be copied !!! */
      new_spec->buffer = thread_data->buffer;
      new_spec->param = (void *) &(new_spec->buffer);
    }
    else {
      new_spec->param = thread_data->param;
    }

    _a0FreeSpecificData(thread_data);
    thread_data = new_spec;
  }
  else {
    a0Debug(("No reusable specific %X found for %d\n", thread_data, prio));
    thread_data->Prio = prio;
    /* Init the thread join-semaphore with 0: join will block */
    if (ret = _a0NewSemaphore(&thread_data->join, 0)) return ret;

#endif /* A0_REUSETHR */

    /* Mutex protecting the liberation of a thread specific structure.
       When a thread exits and it's detached, it frees its structure 
       in _a0ExitThread(). When a thread is not detached, the structure
       must be kept until the _a0JoinThread(). But if _a0DetachThread()
       is called after the thread call _a0ExitThread(), the detach must
       free the structure. All these functions must decide if any of the
       others have executed and a mutual exclusion must exist. */

    if (ret = _a0NewMutex(&thread_data->detach)) return ret;

#ifdef A0_REUSETHR

    /* remember this thread's priority */
    thread_data->Prio = prio;

    /* build a new semaphore with 0: first P will block. */
    _a0NewSemaphore(&thread_data->reuse, 0);
  }

#endif /* A0_REUSETHR */

  /* put specific in thread descriptor to detach or join */
  thread->specific = thread_data;

  /* keep only useful bits in the flags */
  thread_data->flags &= (A0SpecificUsed | A0SpecificBuffer);

#ifdef A0_TRACE
  /* We assign a logical locally unique number for identifying the thread */
  akTraceNewThreadId(&thread->logical_id);
  thread_data->logical_id = thread->logical_id;
#endif

  a0Debug(("effective start spec=%X fcn=%X prm=%X flg=%d "
		"bsz=%d bps=%d\n",
		thread_data, thread_data->function, thread_data->param,
		thread_data->flags, thread_data->buffer.size,
		thread_data->buffer.position));

#ifdef A0_REUSETHR

  if (thread_data->flags & A0SpecificUsed)
    /* there is a thread blocked, waiting to run again */
    /* unlock the semaphore: it will restart the thread */
    return _a0VSemaphore(&thread_data->reuse);
  else

#endif /* A0_REUSETHR */

    return _a0NewThread(thread, pth_sch, pth_pri, stack, _a0ThreadLauncher,
  			(void *) thread_data);
}


/*************************************************************************
	_a0NewThread - starts a new thread with specific POSIX values
*************************************************************************/

extern a0tError _a0NewThread(a0tThread *thread, int sched,
			int prio, size_t stack,
			a0tFunction function, void *arg) {

/* Reminder: the priority passed here is a pthread priority. If this
 * function is to be used somehow by an Athapascan program, the athapascan
 * priority should be converted first to a pthread priority.
 * (see _a0NewSlave for 'user' Athapascan calls)
 */

  a0tThreadAttr attr;
  int ret;
  
  

#if !defined(A0_TSOL2) && !defined(A0_TLIN2) \
    && !defined(A0_TSGI6) && !defined(A0_TAIX4)
  _a0NewThreadAttr(&attr, sched, prio, stack);
#endif


  /* We will have a new thread */
  a0ThreadsTotal++;

#if defined(A0_TSOL2) || defined(A0_TLIN2) \
   || defined(A0_TSGI6) || defined(A0_TAIX4)
  akpthread_attr_init(&attr);
  akpthread_attr_setscope(&attr, akPTHREAD_SCOPE_SYSTEM);
#endif

#ifdef A0_TAIX4
  if (akpthread_attr_setdetachstate(&attr, akPTHREAD_CREATE_JOINABLE))
	_a0Error(A0ErrThreadAttr, "detach");
#endif

  if (ret=akpthread_create(&thread->pth_id, &attr,
		(void *(*)(void *)) function, arg)) {	
	return A0ErrNoMem;
   }
   
   
  akpthread_attr_destroy(&attr);

  return A0ErrOk;
}


/*************************************************************************
	_a0JoinThread - joins with a forked thread
*************************************************************************/

extern a0tError _a0JoinThread(a0tThread *thread, void **result) {

  a0tError ret = 0;
  void *dumb;

a0Debug(("entering on join code...\n"));

  /* just checking */
  if (!thread->specific) return A0ErrInvalid;

#ifdef A0_REUSETHR
  _a0PSemaphore(&thread->specific->join);
#else

a0Debug(("Doing pthread join...\n"));
  a0ThreadsBlockedLock++; /* We may block on the join */
  if (akpthread_join(thread->pth_id, &dumb)) ret = A0ErrInvalid;
  a0ThreadsBlockedLock--; /* ARGH! */
a0Debug(("After pthread join...\n"));
  if (ret) return ret;
#endif

  /* get the result */
  if (result) *result = thread->specific->result;

  /* No need to lock the mutex get the flags.
     The ->join semaphore guarrantees that this code is executed
     AFTER _a0ThreadExit() */

   a0Debug(("Waiting to join thread\n"));
  if (thread->specific->flags & A0SpecificExited) {
  /* Exit passed before the join. Must free the structure here. */
#ifdef A0_REUSETHR
    a0Debug(("Reuse specific %X for %d (join)\n",
		thread->specific, thread->specific->Prio));
    _a0FreeSpecificPrio(thread->specific, thread->specific->Prio);
#else
    _a0FreeSpecificData(thread->specific);
#endif
  }

  return A0ErrOk;
}

  
/*************************************************************************
	_a0DetachThread - detachs a forked thread (join no longer)
*************************************************************************/

extern a0tError _a0DetachThread(a0tThread *thread) {

  int flags;
  a0tError ret;

  /* lock the mutex to mark the detach and get the rest */
  if (ret = _a0LockMutex(&thread->specific->detach)) return ret;
  flags = thread->specific->flags;
  thread->specific->flags |= A0SpecificDetached;
  if (ret = _a0UnlockMutex(&thread->specific->detach)) return ret;

  if (flags & A0SpecificExited) {
  /* Exit passed before the detach. Must free the structure here. */
#ifdef A0_REUSETHR
    a0Debug(("Reuse specific %X for %d (detach)\n",
		thread->specific, thread->specific->Prio));
    _a0FreeSpecificPrio(thread->specific, thread->specific->Prio);
#else
    _a0FreeSpecificData(thread->specific);
#endif
  }
  thread->specific = 0;

#ifndef A0_REUSETHR
  if (akpthread_detach(thread->pth_id)) return A0ErrInvalid;
#endif

  return A0ErrOk;
}


/*************************************************************************
	_a0ExitThread - exits a thread
*************************************************************************/

extern a0tError _a0ExitThread(void *value) {

  a0tThreadSpecific *data;
  a0tError ret;
  int flags;

  _a0GetSpecificData(&data);

  a0Debug(("_a0ExitThread: exiting %08X flags %x\n",
		data, data->flags));

  /* Free the thread's automatic allocated buffer if there's one */
  if (data->flags & A0SpecificBuffer)
    _a0DisposeBuffer(&data->buffer);

  /* lock the mutex to mark the exit and get the rest */
  if (ret = _a0LockMutex(&data->detach)) return ret;
  flags = data->flags;
  data->flags |= A0SpecificExited;
  if (ret = _a0UnlockMutex(&data->detach)) return ret;

#ifdef A0_TRACE
  /* TRACE: insert here callback for thread termination.
   * The logical thread id of the terminatin thread is in
   * data->thread.logical_id
   */
  a0_perf_end_of_user_thread ();
#endif /* A0_TRACE */

  /* One thread less */
  a0ThreadsTotal--;

  if (flags & A0SpecificDetached) {
  /* Detach passed before the exit. Must free the structure here. */
#ifdef A0_REUSETHR
    a0Debug(("Reuse specific %X for %d (exit)\n", data, data->Prio));
    _a0FreeSpecificPrio(data, data->Prio);
#else
    _a0FreeSpecificData(data);
#endif
  }
  else { /* not detached yet */
    data->result = value;
#ifdef A0_REUSETHR
    _a0VSemaphore(&data->join);
#endif
  }

#ifdef A0_REUSETHR
  _a0PSemaphore(&data->reuse);
  a0Debug(("_a0ThreadExit: long jump (%08X)\n", data));
  longjmp(data->begin, 1);
#else
  akpthread_exit(value);
#endif

  _a0Error(A0ErrFatal, "thread exit");
}


/*************************************************************************
	_a0YieldThread - The calling thread does a thread_yield()
*************************************************************************/

extern a0tError _a0YieldThread( void ) {

  /* This call must be here because we don't have timers. The user must
     stop executing his/her thread form time to time to give the kernel
     a chance to reschedule the communication. (argh!) */
  akAdvance();

  aksched_yield();

  return 0;
}


/*************************************************************************
	_a0GetSelfThread - returns the thread descriptor of curr.thread
*************************************************************************/

extern a0tError _a0GetSelfThread(a0tThread *thread) {

  thread->pth_id = akpthread_self();
  _a0GetSpecificData(&thread->specific);

#ifdef A0_TRACE
  akTraceGetThreadId(&thread->logical_id);
#endif /* A0_TRACE */

  return A0ErrOk;
}



