/*
 *
 * _a0urgentd.c : Functions for the URGENT daemon.
 *
 * The URGENT daemon is the one responsible for doing remote requests.
 *    (function executed remotely from within the URGENTD daemon, with no
 *     thread created)
 *
 * by Ilan Ginzburg
 * Apache Project, IMAG, Grenoble, France
 * Jun 19, 1996
 *
 */


/* Athapascan0 source code */
#define A0_SOURCE

#include <akernel.h>
#include "a0error.h"
#include "a0urgentd.h"
#include "a0services.h"
#include "a0thread.h"
#include "a0mpi.h"
#include "a0pt2pt.h"


/* debugging macro */
#define a0Debug_on(a) printf a
#define a0Debug(a)


/*****************************************************************************/

/* Port used for communications with the URGENT daemon */
a0tPort _a0UrgentdPort;

/* Daemon thread id */
static a0tThread _a0UrgentDaemonThread;

static _a0UrgentDaemonRun = 1;


/*****************************************************************************/
/* This is the URGENT daemon thread. It handles Remote Requests. */

static int _a0UrgentDaemonFunction(void *arg) {

  a0tRequest request;
  a0tBuffer RRbuffer; /* Buffer to pass to the RR service function */
  int service_number, dummy_int;
  long dummy_long;
  a0tRemoteFunction RRCode;
  a0tError rc; /* Return code of the RR function */
  a0tRequestData req_data;

  a0Debug(("Start of _a0UrgentDaemonFunction\n"));

#ifdef A0_TRACE
  {
    unsigned long id;
    /* find a trace id and set it */
    akTraceNewThreadId(&id);
    akTraceSetThreadId(id);
  }
#endif /* A0_TRACE */

  while (_a0UrgentDaemonRun) {
    /* Do a blocking probe to get an incoming request.
     * We do a probe so we can allocate the buffer to do the receive on,
     * as remote thread creation buffers can have an arbitrary size.
     */
    /* We do a horrible gororoba to compensate for thread accounting...
     * (we don't want the accounting to reflect athapascan's daemons
     * activity)
     */
    a0ThreadsBlockedComm--; /* gororoba part 1/2 */
    _a0Probe(&_a0UrgentdPort, A0AnySource, 0, &request);
    a0ThreadsBlockedComm++; /* gororoba part 2/2 */

    /* Get the size of the arriving message */
    _a0GetRequestData(&request, &req_data);

    /* Allocate a RR buffer of the correct size. (Actually the allocated
     * buffer is greater than necessary, because size includes the size of
     * the header of the buffer, and in _a0NewBuffer we allocate size bytes +
     * the size of the header. But who cares?... )
     */
    if (_a0NewBuffer(&RRbuffer, A0UrgentBufferType, req_data.size))
    	_a0Error(A0ErrMalloc, "request buff");
	
    a0Debug(("URGENT Daemon %d will read request\n", akNode));

    /* Get the request in the buffer (hopefully the probe worked...) */
    akRecv(RRbuffer.storage, RRbuffer.size,
                MPI_PACKED, req_data.source, _a0UrgentdPort.basetag,
		_a0Communicators[_a0UrgentdPort.comm], &request);

    /* The buffer has just been received, nothing unpacked, so the position
     * is at the beginning of the buffer.
     */
    RRbuffer.position=0; /* As you see, a0tBuffer is not that opaque... :-) */

    /* Do all the necessary unpacking and call the service... */
    _a0Unpack(&RRbuffer, a0_INT, &service_number, 1);

    a0Debug(("URGENT received by %d sended by %d: serv# %d\n", 
            akNode, req_data.source, service_number));

    if (service_number == A0TerminationService) akpthread_exit(0);

    /* Now the buffer is ready to be passed to the service thread, because
     * everything packed by athapascan has been unpacked.
     */

    /* Get the function to call. */
    if (_a0LookupRRService(service_number, &RRCode)) {
       a0Debug(("BAD SERVICE <%d> received by %d sended by %d: RRCode# %d\n", 
              service_number, akNode, req_data.source, RRCode));
       _a0Error(A0ErrInvService, "RR");
    }
    
    a0Debug(("After _a0LookupRRService: service_num=%d service=%X\n",
    					service_number, RRCode));

#ifdef A0_TRACE
    /* TRACE: Insert here callback for RTC remote request start.
     * the service number executed by the request is in service_number
     */
    a0_perf_begin_remote_request (service_number, &request);
#endif /* A0_TRACE */

    /* We call the service function, to which we pass the received buffer */
    rc = RRCode(&RRbuffer);

#ifdef A0_TRACE
    a0_perf_end_remote_request ();
#endif /* A0_TRACE */

    /* What can we do with that return code?... */

    /* The function finished, we free the buffer.
     * WARNING: We may want the buffer to live longer than the Remote Request
     * function. If this is the case, not only we shouldn't free it now, but
     * we should also allocate the a0tBuffer differently... :-)
     */
    _a0DisposeBuffer(&RRbuffer);
  }
}

/*****************************************************************************/

/* This function does the initializations needed for the URGENT daemon, creates
 * the URGENT daemon thread and returns.
 * Upon return from this function, all that's needed by the threads
 * on this node to adress URGENT daemons on other nodes is initialized (as well
 * as all that's needed by the URGENT daemon to answer requests from
 * other nodes).
 */
a0tError _a0InitUrgent()
{
  int rc;

  /* Global creation of a port for communicating with the URGENT daemon */
  _a0NewPort(&_a0UrgentdPort, 0, FALSE);

  /* Gororoba hack... We decrease the total number of threads because
   * _a0NewThread will increase it, and we don't want to include Athapscan
   * threads in this counter.
   */
  a0ThreadsTotal--; /* Sorry 'bout that! */

  /* Creation of the URGENT daemon thread.  */
  rc = _a0NewThread(&_a0UrgentDaemonThread,
    		 _a0DaemonScheduling,
		 _a0PriDaemonHigh,
		 _a0DefaultStack,
		 _a0UrgentDaemonFunction,
		 (void *) NULL);

  a0Debug(("_a0NewThread URGENT returned %d\n", rc));
}

/*****************************************************************************/

a0tError _a0TermUrgent()
{
	void *res;

	_a0UrgentDaemonRun = 0;
	a0Debug(("....Before urgent  URGENT ENDING\n"));
	_a0StartRemoteUrgent(akNode, A0TerminationService, 0, 0);
	a0Debug(("....Esperando o URGENTE terminar!!!!\n"));
	/*akpthread_join(_a0UrgentDaemonThread.pth_id, &res);*/
}
