/*
 * ppong.c: Executes ping-pong using MPI, akernel and
 *          athapascan-0b calls. Measures latency and
 *          Bandwidht.
 *
 * by Alexandre Carissimi
 * Apache Project, IMAG, Grenoble, France
 * January 6,1998
 *
 * Modifications:
 *   20.03.98: ping algorithm included
 */

#include	<a0b.h>
#include	<mpi.h>
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<unistd.h>
#include	<stdarg.h>
#ifdef          TESTAIX
#include        <pthread.h>
#endif

#define MPI

#define		REPS		10
#define		MILISEC		1000.0
#define		SECONDS         1000000.0

int ArgFindName(int, char **, char *);
int ArgGetStr(int, char **, char *, char **);
int ArgGetInt(int, char **, char *, int *);
int ArGetIntList(int, char **, char *, int, int *);


/*
 * This function is defined to exploits particular calls existing
 * on some systems. To compatibility reasons the times is always
 * converted on microseconds.
 */
 
double GetTime(void)
{
#ifdef OTHERS
   struct timeval time;
   double Time;

   gettimeofday(&time, (struct timezone *) NULL);
   Time = ((double) time.tv_sec)*1000000.0 + time.tv_usec;
   return(Time);
#else
   hrtime_t Time;

   Time= gethrtime();
   return((double)(Time/1000.0));
#endif
}

/*
 * This function set executes the ping-pong algorithm using MPI,
 * akernel calls, and a0 calls. All communications are synchronous
 */

#ifdef MPI
double StandardRoundTrip(int rank, int next, int prev,int MsgLen, int Reps)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     MPI_Status status;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        MPI_Recv(Buffer, 1, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &status);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++) {           
           MPI_Send(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD);
           MPI_Recv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &status);
        }
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       MPI_Send(Buffer, 1, MPI_BYTE, next, 0, MPI_COMM_WORLD);            
       for(j=0;j<Reps;j++) {          
          MPI_Recv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &status);
          MPI_Send(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD);
       }
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif

#ifdef AKERNEL
double StandardRoundTrip(int rank, int next, int prev,int MsgLen, int Reps)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     aktHandle  handle;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        akRecv(Buffer, 1, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &handle);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++) {           
           akSend(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD, &handle);
           akRecv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &handle);
        }
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       akSend(Buffer, 1, MPI_BYTE, next, 0, MPI_COMM_WORLD, &handle);            
       for(j=0;j<Reps;j++) {          
          akRecv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &handle);
          akSend(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD, &handle);
       }
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif

#ifdef ATH0
double StandardRoundTrip(int rank,int next,int prev,int MsgLen,int Reps,
                            a0tPort *Port)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     a0tRequest req;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        a0Receive(Port, prev, 0, &req, a0_BYTE, Buffer, 1);         
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++) {    
           a0Send(Port, next, 0, &req, a0_BYTE, Buffer, MsgLen);       
           a0Receive(Port, prev, 0, &req, a0_BYTE, Buffer, MsgLen);
        }
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       a0Send(Port, next, 0, &req, a0_BYTE, Buffer, 1);
       for(j=0;j<Reps;j++) { 
          a0Receive(Port, prev, 0, &req, a0_BYTE, Buffer, MsgLen);
          a0Send(Port, next, 0, &req, a0_BYTE, Buffer, MsgLen);
       }
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif

/*
 * This function set executes the ping algorithm using MPI, akernel
 * calls, and a0 calls. All communications are synchronous
 */

#ifdef MPI
double StandardOneWay(int rank, int next, int prev,int MsgLen, int Reps)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     MPI_Status status;
     MPI_Request req2, req3;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }

      
     if (rank==0) {
        MPI_Recv(Buffer, 1, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &status);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++)           
           MPI_Send(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD);
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       MPI_Send(Buffer, 1, MPI_BYTE, next, 0, MPI_COMM_WORLD);            
       for(j=0;j<Reps;j++)          
          MPI_Recv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &status);
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif

#ifdef AKERNEL
double StandardOneWay(int rank, int next, int prev,int MsgLen, int Reps)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     aktHandle  handle;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        akRecv(Buffer, 1, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &handle);          
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++)            
           akSend(Buffer, MsgLen, MPI_BYTE, next, 0, MPI_COMM_WORLD, &handle);
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       akSend(Buffer, 1, MPI_BYTE, next, 0, MPI_COMM_WORLD, &handle);            
       for(j=0;j<Reps;j++)          
          akRecv(Buffer, MsgLen, MPI_BYTE, prev, 0, MPI_COMM_WORLD, &handle);
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif


#ifdef ATH0
double StandardOneWay(int rank,int next,int prev,int MsgLen,int Reps,
                            a0tPort *Port)
{
     int        i, j;
     char       *Buffer;
     double     InitialTime, EndTime, ElapsedTime;
     a0tRequest req;

     Buffer = (char *) malloc(sizeof(char)*(MsgLen?MsgLen:1));
     if (Buffer == NULL) {
         printf("ABORTING: Memory allocation error\n");
         exit(0);
     }
      
     if (rank==0) {
        a0Receive(Port, prev, 0, &req, a0_BYTE, Buffer, 1);         
        InitialTime = GetTime();  
        for(j=0;j<Reps;j++)     
           a0Send(Port, next, 0, &req, a0_BYTE, Buffer, MsgLen);       
        EndTime = GetTime();
        ElapsedTime = (EndTime - InitialTime)/Reps;
     }
     
    if (rank!=0) {
       a0Send(Port, next, 0, &req, a0_BYTE, Buffer, 1);
       for(j=0;j<Reps;j++)  
          a0Receive(Port, prev, 0, &req, a0_BYTE, Buffer, MsgLen);
    }
    
    free(Buffer);
    return(ElapsedTime);
}
#endif


#ifdef TESTAIX
void *RecNever(void) {
	
int MsgLen=10,prev=0;
char Buffer[10];
MPI_Status Status;

MPI_Recv(Buffer, MsgLen, MPI_BYTE, prev,12, MPI_COMM_WORLD,&Status);
}
#endif


/******************************************************************
 *                                                                *
 ******************************************************************/

       
int main(int argc, char **argv) {

int    Reps=100,          /*times which each message is sended      */
       MsgLen=0,          /*message size (initial equal zero)       */
       FinalSize=65536,   /*final message size                      */
       Stride=1024,       /*increment value for message size        */
       Nt=1,		  /*number of ping-pong threads             */
       Plot=0;            /*control flag to generate results file   */
FILE   *fp=stdout;        /*file pointer (file results)             */
char   *file="none",      /*file's name                             */ 
       *TestName,         /*test to be done                         */
       *ScaleName;        /*scale type message size: byte,KB,or MB  */
double (*ftest)();        /*pointer to test function to be done     */ 

int     rank,size,prev,next; 
int     i, aux[3],log; 

double  RoundTrip,Latency,BandWidth;	
double	ScaleX,ScaleY,Average,a,b;
a0tPort Port;


#ifdef ATH0
  a0Init(&argc, &argv);
  a0NewPort(&Port,4,A0GlobalCreation);
  a0InitCommit();
  rank=a0SelfNode;
  size=a0NodeCount;
#endif

#ifdef AKERNEL
  akInit(&argc, &argv);
  rank=akNode;
  size=akSize;
#endif

#ifdef MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
#endif

 
 if (size < 2) {
     printf("ERROR: It must have more than 1 node to run!!\n");
     exit();
 }
 
 next = (rank+1)%size;
 prev = (rank+size-1) % size;
 

/*
 * Get arguments from the command line. 
 */
 
  if (ArgGetInt(argc, argv, "-reps", &Reps)==-1) {
      printf("#Missing number of repetitions: -reps x \n");
      exit();
  }
   
  if (ArgGetInt(argc, argv, "-nt", &Nt)==-1) {
      printf("#Missing number of threads on aix system: -nt x \n");
      exit();
  }
   
  log = 0;   
  if (ArgFindName(argc, argv, "-log")!=-1) {
     /*FinalSize = 1048576<<2;*/
     FinalSize = 1048576<<2;
     log = 1;
  }
  
  switch(ArgGetIntList(argc, argv, "-size", 3, aux)) {
     case -1:
             printf("#Missing correct size parameters: -size initial final stride \n");
             exit();
     case  1:
             MsgLen= aux[0];
             FinalSize = aux[1];
             Stride = aux[2];
             break;
     default:
             break;
  }
      
  if (FinalSize < 1024) {ScaleX = 1.0;ScaleName="Byte";}
  else if (FinalSize < 1048576) {ScaleX = 1024.0;ScaleName="KByte";}
  else {ScaleX = 1048576.0;ScaleName="MByte";}
  
    
  if (ArgFindName(argc, argv, "-ping")!=-1) {
     ftest = StandardOneWay;
     TestName = "Ping blocking mode"; ScaleY = 1.0;
  }
  else {
      ftest = StandardRoundTrip;
      TestName = "Ping Pong Standard mode"; ScaleY = 2.0;
  } 
 
 
  fp = stdout;
  if (rank == 0 )
     if (Plot=ArgGetStr(argc, argv, "-plot", &file)) {
	fp = fopen(file,"a");
	if (fp==NULL) {
	   printf("ERROR: File creation!! \n");
	   exit();
	}
	fprintf(fp,"#\n#         %s\n#\n", TestName);
        fprintf(fp,"#\n# File: %s \n#\n", file);
	fprintf(fp,"# Scale to message size is: %s \n", ScaleName);
        fprintf(fp,"# Msg. Size    One-way(usecs)");
        fprintf(fp,"  RoundTrip(usecs)  BandWidth (MB/s)\n");
     }
 
   if (rank==0) {
       printf("#\n#         %s\n#\n", TestName);
       printf("# File: %s \n#\n",file);
       printf("# Scale to message size is: %s \n", ScaleName);
       printf("# Msg. Size    One-way(usecs)");
       printf("  RoundTrip(usecs)  BandWidth (MB/s)\n");
   }


#ifdef TESTAIX
   pthread_attr_init(&attr);
   Thread=(pthread_t *)malloc(sizeof(pthread_t)*Nx);
   for(i=0;i<Nx;i++) 
      pthread_create(&Thread[i], &attr, RecNever, (void *)NULL);
  for(i=0;i<Nx;i++)
     pthread_yield();
#endif

       
   while (MsgLen <= FinalSize) { 
      Average=0;
      for(i=0; i<REPS; i++)
          Average+=(*ftest)(rank,next,prev,MsgLen,Reps,&Port);
      if (rank==0) {
          RoundTrip = Average/(double)REPS;
          if (MsgLen==0)
             Latency = RoundTrip/ScaleY;
          BandWidth = ((double)MsgLen*ScaleY)/RoundTrip; /* Bytes per microsec  */
	  BandWidth = BandWidth/1048576.0;               /* converts to MBytes  */
	  BandWidth = BandWidth*(double)SECONDS;         /* converts to seconds */
	  
          printf(" %10.5f   %15.7f  %15.7f     %7.3f \n", (double)MsgLen/ScaleX, 
                 RoundTrip/ScaleY,RoundTrip*2.0/ScaleY, BandWidth);
          if (Plot)
             fprintf(fp," %10.5f   %15.7f  %15.7f     %7.3f \n", 
	            (double)MsgLen/ScaleX,RoundTrip/ScaleY,RoundTrip*2.0/ScaleY,
		    BandWidth);
          }  
          
       if (log) 
          MsgLen = MsgLen?MsgLen<<2:1;
       else        
          MsgLen+=Stride;
   }

   if (rank==0) {
         printf("#\n#Latency evaluated: %-6.2f us \n", Latency);
      if (Plot) 
         fprintf(fp,"#\n#Latency evaluated: %-6.2f us \n", Latency);      
   }


#ifdef ATH0
  a0Terminate();
#endif

#ifdef AKERNEL
  akFinalize();
  fflush(stdout);
#endif

#ifdef MPI
  MPI_Finalize();
#endif

   fclose(fp);
}
